# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from pathlib import Path
from types import MethodType

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from covalent._shared_files import logger

from .base import BaseTrigger

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class DirEventHandler(FileSystemEventHandler):
    def __init__(self) -> None:
        self.supported_event_to_func_names = {
            "created": "on_created",
            "deleted": "on_deleted",
            "modified": "on_modified",
            "moved": "on_moved",
            "closed": "on_closed",
        }


class DirTrigger(BaseTrigger):
    """
    Directory or File based trigger which watches for events in said file/dir
    and performs a trigger action whenever they happen.

    Args:
        dir_path: Path to the file/dir which is to be observed for events
        event_names: List of event names on which to perform the trigger action.
                     Possible options can be a subset of: `["created", "deleted", "modified", "moved", "closed"]`.
        batch_size: The number of changes to wait for before performing the trigger action, default is 1.
        recursive: Whether to recursively watch the directory, default is False.

    Attributes:
        self.dir_path: Path to the file/dir which is to be observed for events
        self.event_names: List of event names on which to perform the trigger action.
                          Possible options can be a subset of: `["created", "deleted", "modified", "moved", "closed"]`
        self.batch_size: The number of events to wait for before performing the trigger action, default is `1`.
        self.recursive: Whether to recursively watch the directory, default is False.
        self.n_changes: Number of events since last trigger action. Whenever `self.n_changes == self.batch_size` a trigger action happens.
    """

    def __init__(
        self,
        dir_path,
        event_names,
        batch_size: int = 1,
        recursive: bool = False,
        lattice_dispatch_id: str = None,
        dispatcher_addr: str = None,
        triggers_server_addr: str = None,
    ):
        super().__init__(lattice_dispatch_id, dispatcher_addr, triggers_server_addr)

        self.dir_path = dir_path

        if isinstance(event_names, str):
            event_names = [event_names]
        self.event_names = event_names

        self.batch_size = batch_size
        self.recursive = recursive

        self.observe_blocks = False
        self.event_handler = None

    def attach_methods_to_handler(self) -> None:
        """
        Dynamically attaches and overrides the "on_*" methods to the handler
        depending on which ones are requested by the user.

        Args:
            event_names: List of event names upon which to perform a trigger action
        """

        app_log.warning("Attaching methods to dir handler")

        self.n_changes = 0

        def proxy_trigger(_, event_object):
            self.n_changes += 1
            if self.n_changes == self.batch_size:
                self.trigger()
                self.n_changes = 0

        for en in self.event_names:
            func_name = self.event_handler.supported_event_to_func_names[en]
            proxy_trigger.__name__ = func_name
            setattr(self.event_handler, func_name, MethodType(proxy_trigger, self.event_handler))

    def observe(self) -> None:
        """
        Start observing the file/dir for any possible events among the ones
        mentioned in `self.event_names`.
        Currently only supports running within the Covalent/Triggers server.
        """

        app_log.warning(f"In DirTrigger's observe, dir path is: {self.dir_path}")

        # Resolving the path at the place where observation will happen
        self.dir_path = str(Path(self.dir_path).expanduser().resolve())

        self.event_loop = asyncio.get_running_loop()

        self.event_handler = DirEventHandler()

        # Attach methods before scheduling the observer
        self.attach_methods_to_handler()

        self.observer = Observer()
        print("Schedule is: ", type(self.observer.schedule))
        self.observer.schedule(self.event_handler, self.dir_path, recursive=self.recursive)
        self.observer.start()

    def stop(self) -> None:
        """
        Stop observing the file or directory for changes.
        """
        self.observer.stop()
        self.observer.join()



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sqlite3
import time
from functools import partial
from threading import Event
from typing import List

from covalent._shared_files import logger

from .base import BaseTrigger

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class SQLiteTrigger(BaseTrigger):
    """
    SQLite based Trigger which can read for changes in a SQLite database
    and trigger workflows based on that.

    Args:
        db_path: Absolute path to the database file
        table_name: Name of the table to observe
        poll_interval: Time in seconds to wait for before reading the database again
        where_clauses: List of "WHERE" conditions, e.g. ["id > 2", "status = pending"], to check when
                       polling the database
        trigger_after_n: Number of times the event must happen after which the workflow will be triggered.
                         e.g value of 2 means workflow will be triggered once the event has occurred twice.
        lattice_dispatch_id: Lattice dispatch id of the workflow to be triggered
        dispatcher_addr: Address of the dispatcher server
        triggers_server_addr: Address of the triggers server

    Attributes:
        self.db_path: Absolute path to the database file
        self.table_name: Name of the table to observe
        self.poll_interval: Time in seconds to wait for before reading the database again
        self.where_clauses: List of "WHERE" conditions, e.g. ["id > 2", "status = pending"], to check when
                            polling the database
        self.trigger_after_n: Number of times the event must happen after which the workflow will be triggered.
                              e.g value of 2 means workflow will be triggered once the event has occurred twice.
        self.stop_flag: Thread safe flag used to check whether the stop condition has been met

    """

    def __init__(
        self,
        db_path: str,
        table_name: str,
        poll_interval: int = 1,
        where_clauses: List[str] = None,
        trigger_after_n: int = 1,
        lattice_dispatch_id: str = None,
        dispatcher_addr: str = None,
        triggers_server_addr: str = None,
    ):
        super().__init__(lattice_dispatch_id, dispatcher_addr, triggers_server_addr)

        self.db_path = db_path
        self.table_name = table_name
        self.poll_interval = poll_interval

        self.where_clauses = where_clauses

        self.trigger_after_n = trigger_after_n

        self.stop_flag = None

    def observe(self) -> None:
        """
        Keep performing the trigger action as long as
        where conditions are met or until stop has being called
        """

        app_log.debug("Inside SQLiteTrigger's observe")
        event_count = 0

        connection = sqlite3.connect(self.db_path)
        connection.row_factory = sqlite3.Row

        cursor = connection.cursor()

        sql_poll_cmd = f"SELECT * FROM {self.table_name}"

        if self.where_clauses:
            sql_poll_cmd += " WHERE "
            sql_poll_cmd += " AND ".join(list(self.where_clauses))

        execute_cmd = partial(cursor.execute, sql_poll_cmd)

        self.stop_flag = Event()
        while not self.stop_flag.is_set():
            # Read the DB with specified command
            try:
                execute_cmd()
            except sqlite3.OperationalError:
                time.sleep(self.poll_interval)
                continue

            # If command ran successfuly, trigger the workflow
            if selected_rows := cursor.fetchall():
                event_count += 1
                if event_count == self.trigger_after_n:
                    self.trigger()
                    event_count = 0

            time.sleep(self.poll_interval)

        cursor.close()
        connection.close()

    def stop(self) -> None:
        """
        Stop the running `self.observe()` method by setting the `self.stop_flag` flag.
        """

        self.stop_flag.set()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Server-side file transfer utilities
"""

from concurrent.futures import ThreadPoolExecutor

from covalent._file_transfer import FileTransfer
from covalent._shared_files import logger

app_log = logger.app_log
am_pool = ThreadPoolExecutor()


def cp(src_uri: str, dest_uri: str, transfer_options: dict = {}):
    ft = FileTransfer(src_uri, dest_uri)
    pre_hook, transfer_callable = FileTransfer(src_uri, dest_uri).cp()
    transfer_callable()



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Implement several different serialization methods for QNode output data written
to the database.
"""
import warnings
import zlib
from abc import ABC, abstractmethod
from enum import Enum
from typing import Sequence, Union

import cloudpickle as pickle
import lmdb
import orjson
from lmdbm.lmdbm import Lmdb, remove_lmdbm


class _Serializer(ABC):
    """
    base class for serializer strategies
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """
        strategy name
        """

    @abstractmethod
    def pre_value(self, value):
        """
        returns processed value to be written to db
        """
        raise NotImplementedError()

    @abstractmethod
    def post_value(self, value):
        """
        post-process value read from db and return
        """
        raise NotImplementedError()


class _OrjsonStrategy(_Serializer):
    """
    uses `orjson` and `zlib` to serialize/deserialize
    """

    name = "orjson"

    def pre_value(self, value):
        return zlib.compress(orjson.dumps(value, option=orjson.OPT_SERIALIZE_NUMPY))

    def post_value(self, value):
        return orjson.loads(zlib.decompress(value))


class _PickleStrategy(_Serializer):
    """
    uses `cloudpickle` and `zlib` to serialize/deserialize
    """

    name = "pickle"

    def pre_value(self, value):
        return zlib.compress(pickle.dumps(value))

    def post_value(self, value):
        return pickle.loads(zlib.decompress(value))


class _FallbackStrategy(_Serializer):
    """
    tries multiple strategies until success
    """

    name = "fallback"

    def __init__(self, strategies):
        self.strategies = strategies

    def pre_value(self, value):
        for strategy in self.strategies:
            try:
                return strategy.pre_value(value)
            except TypeError as te:
                warnings.warn(
                    f"serialization strategy '{strategy.name}' failed on `pre_value`; "
                    f"value: {value}; error: {te}."
                )
        raise RuntimeError("all strategies failed to encode data")

    def post_value(self, value):
        for strategy in self.strategies:
            try:
                return strategy.post_value(value)
            except TypeError as te:
                warnings.warn(
                    f"serialization strategy '{strategy.name}' failed on `post_value`; "
                    f"value: {value}; error: {te}."
                )
        raise RuntimeError("all strategies failed to decode data")


class Strategy(Enum):
    """
    available serialization strategies
    """

    ORJSON = _OrjsonStrategy
    PICKLE = _PickleStrategy
    FALLBACK = _FallbackStrategy


class JsonLmdb(Lmdb):
    """
    custom `Lmdb` implementation with pre- and post-value strategy option
    """

    def __init__(self, strategy_type: Union[Strategy, Sequence[Strategy]], **kw):
        self._strategy_map = {}
        self.strategy = self.init_strategy(strategy_type)
        super().__init__(**kw)

    def _pre_key(self, key):
        return key.encode("utf-8")

    def _post_key(self, key):
        return key.decode("utf-8")

    def _pre_value(self, value):
        return self.strategy.pre_value(value)

    def _post_value(self, value):
        return self.strategy.post_value(value)

    @property
    def strategy_map(self):
        """
        allows access to strategies by str name
        """
        if not self._strategy_map:
            self._strategy_map = {s.value.name: s.value for s in list(Strategy)}
        return self._strategy_map

    def init_strategy(self, strategy_type) -> _Serializer:
        """
        initialize an instance of the named strategy
        """
        if isinstance(strategy_type, Strategy):
            return self._init_single_strategy(strategy_type)

        # strategy with fallback
        strategies = [self._init_single_strategy(typ) for typ in strategy_type]
        return _FallbackStrategy(strategies)

    def _init_single_strategy(self, strategy_type) -> _Serializer:
        if not isinstance(strategy_type, Strategy):
            raise TypeError(f"expected Strategy, not {type(strategy_type.__class__.__name__)}")

        strategy_cls = self.strategy_map.get(strategy_type.value.name)
        if strategy_cls is None:
            raise ValueError(f"unknown database strategy '{strategy_type}'")
        return strategy_cls()

    @classmethod
    def open_with_strategy(
        cls,
        file,
        flag="r",
        mode=0o755,
        map_size=2**20,
        *,
        strategy_name,
        autogrow=True,
    ):
        """
        Custom open classmethod that takes a (new) strategy argument. Mostly
        replicates original `Lmdb.open`, except passing `strategy_name` to initializer.

        Opens the database `file`.
        `flag`: r (read only, existing), w (read and write, existing),
                c (read, write, create if not exists), n (read, write, overwrite existing)
        `map_size`: Initial database size. Defaults to 2**20 (1MB).
        `autogrow`: Automatically grow the database size when `map_size` is exceeded.
                WARNING: Set this to `False` for multi-process write access.
        `strategy_name`: either 'orjson' or 'pickle'
        """

        if flag == "r":  # Open existing database for reading only (default)
            env = lmdb.open(
                file, map_size=map_size, max_dbs=1, readonly=True, create=False, mode=mode
            )
        elif flag == "w":  # Open existing database for reading and writing
            env = lmdb.open(
                file, map_size=map_size, max_dbs=1, readonly=False, create=False, mode=mode
            )
        elif flag == "c":  # Open database for reading and writing, creating it if it doesn't exist
            env = lmdb.open(
                file, map_size=map_size, max_dbs=1, readonly=False, create=True, mode=mode
            )
        elif flag == "n":  # Always create a new, empty database, open for reading and writing
            remove_lmdbm(file)
            env = lmdb.open(
                file, map_size=map_size, max_dbs=1, readonly=False, create=True, mode=mode
            )
        else:
            raise ValueError("Invalid flag")

        return cls(strategy_name, env=env, autogrow=autogrow)



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to transform URIs"""

import enum
from typing import Optional

from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.utils import format_server_url

SERVER_URL = format_server_url(get_config("dispatcher.address"), get_config("dispatcher.port"))

app_log = logger.app_log


class AssetScope(enum.Enum):
    DISPATCH = "dispatch"
    LATTICE = "lattice"
    NODE = "node"


class URIFilterPolicy(enum.Enum):
    raw = "raw"  # expose raw URIs
    http = "http"  # return data endpoints


def _srv_asset_uri(
    uri: str, attrs: dict, scope: AssetScope, dispatch_id: str, node_id: Optional[int], key: str
) -> str:
    base_uri = f"{SERVER_URL}/api/v2/dispatches/{dispatch_id}"

    if scope == AssetScope.DISPATCH:
        return f"{base_uri}/assets/{key}"
    elif scope == AssetScope.LATTICE:
        return f"{base_uri}/lattice/assets/{key}"
    else:
        return f"{base_uri}/electrons/{node_id}/assets/{key}"


def _raw(
    uri: str, attrs: dict, scope: AssetScope, dispatch_id: str, node_id: Optional[int], key: str
):
    return uri


_filter_map = {
    URIFilterPolicy.raw: _raw,
    URIFilterPolicy.http: _srv_asset_uri,
}


def filter_asset_uri(
    filter_policy: URIFilterPolicy,
    uri: str,
    attrs: dict,
    scope: AssetScope,
    dispatch_id: str,
    node_id: Optional[int],
    key: str,
) -> str:
    """Transform an internal URI for an asset to an external URI.

    Parameters:
        uri: internal URI
        attrs: attributes for the external URI
        scope: asset scope ("dispatch", "lattice", "node")
        key: asset key

    Returns:
        The external URI for the asset

    """

    selected_filter = _filter_map[filter_policy]
    return selected_filter(
        uri=uri,
        attrs=attrs,
        scope=scope,
        dispatch_id=dispatch_id,
        node_id=node_id,
        key=key,
    )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Endpoints for uploading and downloading workflow assets"""

import asyncio
import mmap
import os
from functools import lru_cache
from typing import Tuple, Union

import aiofiles
import aiofiles.os
from fastapi import APIRouter, Header, HTTPException, Request
from fastapi.responses import StreamingResponse
from furl import furl

from covalent._serialize.electron import ASSET_TYPES as ELECTRON_ASSET_TYPES
from covalent._serialize.lattice import ASSET_TYPES as LATTICE_ASSET_TYPES
from covalent._serialize.result import ASSET_TYPES as RESULT_ASSET_TYPES
from covalent._serialize.result import AssetType
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._workflow.transportable_object import TOArchiveUtils

from .._dal.result import get_result_object
from .._db.datastore import workflow_db
from .models import (
    AssetRepresentation,
    DispatchAssetKey,
    ElectronAssetKey,
    LatticeAssetKey,
    range_pattern,
    range_regex,
)

app_log = logger.app_log
log_stack_info = logger.log_stack_info

router: APIRouter = APIRouter()

app_log = logger.app_log
log_stack_info = logger.log_stack_info

router: APIRouter = APIRouter()

_background_tasks = set()

LRU_CACHE_SIZE = get_config("dispatcher.asset_cache_size")


@router.get("/dispatches/{dispatch_id}/electrons/{node_id}/assets/{key}")
def get_node_asset(
    dispatch_id: str,
    node_id: int,
    key: ElectronAssetKey,
    representation: Union[AssetRepresentation, None] = None,
    Range: Union[str, None] = Header(default=None, regex=range_regex),
):
    """Returns an asset for an electron.

    Args:
        dispatch_id: The dispatch's unique id.
        node_id: The id of the electron.
        key: The name of the asset
        representation: (optional) the representation ("string" or "pickle") of a `TransportableObject`
        range: (optional) range request header

    If `representation` is specified, it will override the range request.
    """
    start_byte = 0
    end_byte = -1

    try:
        if Range:
            start_byte, end_byte = _extract_byte_range(Range)

        if end_byte >= 0 and end_byte < start_byte:
            raise HTTPException(
                status_code=400,
                detail="Invalid byte range",
            )
        app_log.debug(
            f"Requested asset {key.value} ([{start_byte}:{end_byte}]) for node {dispatch_id}:{node_id}"
        )

        result_object = get_cached_result_object(dispatch_id)

        app_log.debug(f"LRU cache info: {get_cached_result_object.cache_info()}")

        node = result_object.lattice.transport_graph.get_node(node_id)
        with workflow_db.session() as session:
            asset = node.get_asset(key=key.value, session=session)

        # Explicit representation overrides the byte range
        if representation is None or ELECTRON_ASSET_TYPES[key.value] != AssetType.TRANSPORTABLE:
            start_byte = start_byte
            end_byte = end_byte
        elif representation == AssetRepresentation.string:
            start_byte, end_byte = _get_tobj_string_offsets(asset.internal_uri)
        else:
            start_byte, end_byte = _get_tobj_pickle_offsets(asset.internal_uri)

        app_log.debug(f"Serving byte range {start_byte}:{end_byte} of {asset.internal_uri}")
        generator = _generate_file_slice(asset.internal_uri, start_byte, end_byte)
        return StreamingResponse(generator)

    except Exception as e:
        app_log.debug(e)
        raise


@router.get("/dispatches/{dispatch_id}/assets/{key}")
def get_dispatch_asset(
    dispatch_id: str,
    key: DispatchAssetKey,
    representation: Union[AssetRepresentation, None] = None,
    Range: Union[str, None] = Header(default=None, regex=range_regex),
):
    """Returns a dynamic asset for a workflow

    Args:
        dispatch_id: The dispatch's unique id.
        key: The name of the asset
        representation: (optional) the representation ("string" or "pickle") of a `TransportableObject`
        range: (optional) range request header

    If `representation` is specified, it will override the range request.
    """
    start_byte = 0
    end_byte = -1

    try:
        if Range:
            start_byte, end_byte = _extract_byte_range(Range)

        if end_byte >= 0 and end_byte < start_byte:
            raise HTTPException(
                status_code=400,
                detail="Invalid byte range",
            )
        app_log.debug(
            f"Requested asset {key.value} ([{start_byte}:{end_byte}]) for dispatch {dispatch_id}"
        )

        result_object = get_cached_result_object(dispatch_id)

        app_log.debug(f"LRU cache info: {get_cached_result_object.cache_info()}")
        with workflow_db.session() as session:
            asset = result_object.get_asset(key=key.value, session=session)

        # Explicit representation overrides the byte range
        if representation is None or RESULT_ASSET_TYPES[key.value] != AssetType.TRANSPORTABLE:
            start_byte = start_byte
            end_byte = end_byte
        elif representation == AssetRepresentation.string:
            start_byte, end_byte = _get_tobj_string_offsets(asset.internal_uri)
        else:
            start_byte, end_byte = _get_tobj_pickle_offsets(asset.internal_uri)

        app_log.debug(f"Serving byte range {start_byte}:{end_byte} of {asset.internal_uri}")
        generator = _generate_file_slice(asset.internal_uri, start_byte, end_byte)
        return StreamingResponse(generator)
    except Exception as e:
        app_log.debug(e)
        raise


@router.get("/dispatches/{dispatch_id}/lattice/assets/{key}")
def get_lattice_asset(
    dispatch_id: str,
    key: LatticeAssetKey,
    representation: Union[AssetRepresentation, None] = None,
    Range: Union[str, None] = Header(default=None, regex=range_regex),
):
    """Returns a static asset for a workflow

    Args:
        dispatch_id: The dispatch's unique id.
        key: The name of the asset
        representation: (optional) the representation ("string" or "pickle") of a `TransportableObject`
        range: (optional) range request header

    If `representation` is specified, it will override the range request.
    """
    start_byte = 0
    end_byte = -1

    try:
        if Range:
            start_byte, end_byte = _extract_byte_range(Range)

        if end_byte >= 0 and end_byte < start_byte:
            raise HTTPException(
                status_code=400,
                detail="Invalid byte range",
            )
        app_log.debug(
            f"Requested lattice asset {key.value} ([{start_byte}:{end_byte}])for dispatch {dispatch_id}"
        )

        result_object = get_cached_result_object(dispatch_id)
        app_log.debug(f"LRU cache info: {get_cached_result_object.cache_info()}")

        with workflow_db.session() as session:
            asset = result_object.lattice.get_asset(key=key.value, session=session)

        # Explicit representation overrides the byte range
        if representation is None or LATTICE_ASSET_TYPES[key.value] != AssetType.TRANSPORTABLE:
            start_byte = start_byte
            end_byte = end_byte
        elif representation == AssetRepresentation.string:
            start_byte, end_byte = _get_tobj_string_offsets(asset.internal_uri)
        else:
            start_byte, end_byte = _get_tobj_pickle_offsets(asset.internal_uri)

        app_log.debug(f"Serving byte range {start_byte}:{end_byte} of {asset.internal_uri}")
        generator = _generate_file_slice(asset.internal_uri, start_byte, end_byte)
        return StreamingResponse(generator)

    except Exception as e:
        app_log.debug(e)
        raise e


@router.put("/dispatches/{dispatch_id}/electrons/{node_id}/assets/{key}")
async def upload_node_asset(
    req: Request,
    dispatch_id: str,
    node_id: int,
    key: ElectronAssetKey,
    content_length: int = Header(default=0),
    digest_alg: Union[str, None] = Header(default=None),
    digest: Union[str, None] = Header(default=None),
):
    """Upload an electron asset.

    Args:
        dispatch_id: The dispatch's unique id.
        node_id: The electron id.
        key: The name of the asset
        asset_file: (body) The file to be uploaded
        content_length: (header)
        digest: (header)
    """
    app_log.debug(f"Uploading node asset {dispatch_id}:{node_id}:{key} ({content_length} bytes) ")

    try:
        metadata = {"size": content_length, "digest_alg": digest_alg, "digest": digest}
        internal_uri = await _run_in_executor(
            _update_node_asset_metadata,
            dispatch_id,
            node_id,
            key,
            metadata,
        )
        # Stream the request body to object store
        await _transfer_data(req, internal_uri)

        return f"Uploaded file to {internal_uri}"
    except Exception as e:
        app_log.debug(e)
        raise


@router.put("/dispatches/{dispatch_id}/assets/{key}")
async def upload_dispatch_asset(
    req: Request,
    dispatch_id: str,
    key: DispatchAssetKey,
    content_length: int = Header(default=0),
    digest_alg: Union[str, None] = Header(default=None),
    digest: Union[str, None] = Header(default=None),
):
    """Upload a dispatch asset.

    Args:
        dispatch_id: The dispatch's unique id.
        key: The name of the asset
        asset_file: (body) The file to be uploaded
        content_length: (header)
        digest: (header)
    """
    app_log.debug(f"Uploading dispatch asset {dispatch_id}:{key} ({content_length} bytes) ")
    try:
        metadata = {"size": content_length, "digest_alg": digest_alg, "digest": digest}
        internal_uri = await _run_in_executor(
            _update_dispatch_asset_metadata,
            dispatch_id,
            key,
            metadata,
        )
        # Stream the request body to object store
        await _transfer_data(req, internal_uri)
        return f"Uploaded file to {internal_uri}"
    except Exception as e:
        app_log.debug(e)
        raise


@router.put("/dispatches/{dispatch_id}/lattice/assets/{key}")
async def upload_lattice_asset(
    req: Request,
    dispatch_id: str,
    key: LatticeAssetKey,
    content_length: int = Header(default=0),
    digest_alg: Union[str, None] = Header(default=None),
    digest: Union[str, None] = Header(default=None),
):
    """Upload a lattice asset.

    Args:
        dispatch_id: The dispatch's unique id.
        key: The name of the asset
        asset_file: (body) The file to be uploaded
        content_length: (header)
        digest: (header)
    """
    try:
        app_log.debug(f"Uploading lattice asset {dispatch_id}:{key} ({content_length} bytes) ")
        metadata = {"size": content_length, "digest_alg": digest_alg, "digest": digest}
        internal_uri = await _run_in_executor(
            _update_lattice_asset_metadata,
            dispatch_id,
            key,
            metadata,
        )
        # Stream the request body to object store
        await _transfer_data(req, internal_uri)
        return f"Uploaded file to {internal_uri}"
    except Exception as e:
        app_log.debug(e)
        raise


def _generate_file_slice(file_url: str, start_byte: int, end_byte: int, chunk_size: int = 65536):
    """Generator of a byte slice from a file.

    Args:
        file_url: A file:/// type URL pointing to the file
        start_byte: The beginning of the byte range
        end_byte: The end of the byte range, or -1 to select [start_byte:]
        chunk_size: The size of each chunk

    Returns:
        Yields chunks of size <= chunk_size
    """
    byte_pos = start_byte
    file_path = str(furl(file_url).path)
    with open(file_path, "rb") as f:
        f.seek(start_byte)
        if end_byte < 0:
            for chunk in f:
                yield chunk
        else:
            while byte_pos + chunk_size < end_byte:
                byte_pos += chunk_size
                yield f.read(chunk_size)
            yield f.read(end_byte - byte_pos)


def _extract_byte_range(byte_range_header: str) -> Tuple[int, int]:
    """Extract the byte range from a range request header."""
    start_byte = 0
    end_byte = -1
    match = range_pattern.match(byte_range_header)
    start = match.group(1)
    end = match.group(2)
    start_byte = int(start)
    if end:
        end_byte = int(end)

    return start_byte, end_byte


# Helpers for TransportableObject


def _get_tobj_string_offsets(file_url: str) -> Tuple[int, int]:
    """Get the byte range for the str rep of a stored TObj.

    For a first implementation we just query the filesystem directly.

    Args:
        file_url: A file:/// URL pointing to the TransportableObject

    Returns:
        (start_byte, end_byte)
    """

    file_path = str(furl(file_url).path)
    filelen = os.path.getsize(file_path)
    with open(file_path, "rb+") as f:
        with mmap.mmap(f.fileno(), filelen) as mm:
            # TOArchiveUtils operates on byte arrays
            return TOArchiveUtils.string_byte_range(mm)


def _get_tobj_pickle_offsets(file_url: str) -> Tuple[int, int]:
    """Get the byte range for the picklebytes of a stored TObj.

    For a first implementation we just query the filesystem directly.

    Args:
        file_url: A file:/// URL pointing to the TransportableObject

    Returns:
        (start_byte, -1)
    """

    file_path = str(furl(file_url).path)
    filelen = os.path.getsize(file_path)
    with open(file_path, "rb+") as f:
        with mmap.mmap(f.fileno(), filelen) as mm:
            # TOArchiveUtils operates on byte arrays
            return TOArchiveUtils.data_byte_range(mm)


# This must only be used for static data as we don't have yet any
# intelligent invalidation logic.
@lru_cache(maxsize=LRU_CACHE_SIZE)
def get_cached_result_object(dispatch_id: str):
    try:
        with workflow_db.session() as session:
            srv_res = get_result_object(dispatch_id, bare=False, session=session)
            app_log.debug(f"Caching result {dispatch_id}")

            # Prepopulate asset maps to avoid DB lookups

            srv_res.populate_asset_map(session)
            srv_res.lattice.populate_asset_map(session)

            tg = srv_res.lattice.transport_graph
            g = tg.get_internal_graph_copy()
            for node_id in g.nodes():
                node = tg.get_node(node_id, session)
                node.populate_asset_map(session)
    except KeyError:
        raise HTTPException(
            status_code=404,
            detail=f"The requested dispatch ID {dispatch_id} was not found.",
        )

    return srv_res


def _filter_null_metadata(metadata):
    # Filter out null updates
    return {k: v for k, v in metadata.items() if v is not None}


def _update_node_asset_metadata(dispatch_id, node_id, key, metadata) -> str:
    result_object = get_cached_result_object(dispatch_id)

    app_log.debug(f"LRU cache info: {get_cached_result_object.cache_info()}")
    node = result_object.lattice.transport_graph.get_node(node_id)
    with workflow_db.session() as session:
        asset = node.get_asset(key=key.value, session=session)
        app_log.debug(f"Asset uri {asset.internal_uri}")

        # Update asset metadata
        update = _filter_null_metadata(metadata)
        node.update_assets(updates={key: update}, session=session)
        app_log.debug(f"Updated node asset {dispatch_id}:{node_id}:{key}")

        return asset.internal_uri


def _update_lattice_asset_metadata(dispatch_id, key, metadata) -> str:
    result_object = get_cached_result_object(dispatch_id)

    app_log.debug(f"LRU cache info: {get_cached_result_object.cache_info()}")
    with workflow_db.session() as session:
        asset = result_object.lattice.get_asset(key=key.value, session=session)

        # Update asset metadata
        update = _filter_null_metadata(metadata)
        result_object.lattice.update_assets(updates={key: update}, session=session)
        app_log.debug(f"Updated size for lattice asset {dispatch_id}:{key}")

        return asset.internal_uri


def _update_dispatch_asset_metadata(dispatch_id, key, metadata) -> str:
    result_object = get_cached_result_object(dispatch_id)

    app_log.debug(f"LRU cache info: {get_cached_result_object.cache_info()}")
    with workflow_db.session() as session:
        asset = result_object.get_asset(key=key.value, session=session)

        # Update asset metadata
        update = _filter_null_metadata(metadata)
        result_object.update_assets(updates={key: update}, session=session)
        app_log.debug(f"Updated size for dispatch asset {dispatch_id}:{key}")
        return asset.internal_uri


async def _transfer_data(req: Request, destination_url: str):
    dest_url = furl(destination_url)
    dest_path = str(dest_url.path)

    # Stream data to a temporary file, then replace the destination
    # file atomically
    tmp_path = f"{dest_path}.tmp"
    app_log.debug(f"Streaming file upload to {tmp_path}")

    async with aiofiles.open(tmp_path, "wb") as f:
        async for chunk in req.stream():
            await f.write(chunk)

    await aiofiles.os.replace(tmp_path, dest_path)


def _run_in_executor(function, *args) -> asyncio.Future:
    loop = asyncio.get_running_loop()
    return loop.run_in_executor(None, function, *args)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess
import tempfile
from copy import deepcopy
from typing import List, Union

from .deps import Deps
from .transport import TransportableObject


def apply_pip_deps(pkgs: [] = [], requirements_content: str = ""):
    if requirements_content:
        reqs_filename = ""
        with tempfile.NamedTemporaryFile("w", delete=False) as f:
            f.write(requirements_content)
            reqs_filename = f.name
        cmd = f"pip install --no-input -r {reqs_filename}".split()

    else:
        pkg_list = " ".join(pkgs)
        cmd = f"pip install --no-input {pkg_list}".split()

    subprocess.run(cmd, stdin=subprocess.DEVNULL, check=True, capture_output=True)


class DepsPip(Deps):
    """PyPI packages to be installed before executing an electron

    A specification of Pip packages to be installed

    Attributes:
        packages: A list of PyPI packages to install
        reqs_path: Path to requirements.txt (overrides `packages`)

    These packages are installed in an electron's execution
    environment just before the electron is run.

    """

    def __init__(self, packages: Union[List, str] = [], reqs_path: str = ""):
        if isinstance(packages, str):
            self.packages = [packages]
        else:
            self.packages = packages

        self.reqs_path = reqs_path
        self.requirements_content = ""

        if self.reqs_path:
            with open(self.reqs_path, "r") as f:
                self.requirements_content = f.read()

        apply_args = [self.packages, self.requirements_content]

        super().__init__(apply_fn=apply_pip_deps, apply_args=apply_args)

    def to_dict(self) -> dict:
        """Return a JSON-serializable dictionary representation of self"""
        attributes = self.__dict__.copy()
        for k, v in attributes.items():
            if isinstance(v, TransportableObject):
                attributes[k] = v.to_dict()
        return {"type": "DepsPip", "short_name": self.short_name(), "attributes": attributes}

    def from_dict(self, object_dict) -> "DepsPip":
        """Rehydrate a dictionary representation

        Args:
            object_dict: a dictionary representation returned by `to_dict`

        Returns:
            self

        Instance attributes will be overwritten.
        """

        if not object_dict:
            return self

        attributes = deepcopy(object_dict)["attributes"]
        for k, v in attributes.items():
            if isinstance(v, dict) and v.get("type", None) == "TransportableObject":
                attributes[k] = TransportableObject.from_dict(v)

        self.__dict__ = attributes

        return self



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class implementation of the transport graph in the workflow graph."""

import datetime
import json
from contextlib import contextmanager
from copy import deepcopy
from typing import Any, Callable, Dict

import cloudpickle
import networkx as nx

from .._shared_files.defaults import parameter_prefix
from .._shared_files.util_classes import RESULT_STATUS, Status
from .transportable_object import TransportableObject


# Functions for encoding the transport graph
def encode_metadata(metadata: dict) -> dict:
    # Idempotent
    # Special handling required for: executor, workflow_executor, deps, call_before/after, triggers

    encoded_metadata = deepcopy(metadata)
    if "executor" in metadata:
        if "executor_data" not in metadata:
            encoded_metadata["executor_data"] = None if metadata["executor"] is None else {}
        if metadata["executor"] is not None and not isinstance(metadata["executor"], str):
            encoded_executor = metadata["executor"].to_dict()
            encoded_metadata["executor"] = encoded_executor["short_name"]
            encoded_metadata["executor_data"] = encoded_executor

    if "workflow_executor" in metadata:
        if "workflow_executor_data" not in metadata:
            encoded_metadata["workflow_executor_data"] = (
                None if metadata["workflow_executor"] is None else {}
            )
        if metadata["workflow_executor"] is not None and not isinstance(
            metadata["workflow_executor"], str
        ):
            encoded_wf_executor = metadata["workflow_executor"].to_dict()
            encoded_metadata["workflow_executor"] = encoded_wf_executor["short_name"]
            encoded_metadata["workflow_executor_data"] = encoded_wf_executor

    # Bash Deps, Pip Deps, Env Deps, etc

    if "hooks" in metadata and metadata["hooks"] is not None:
        hooks = metadata["hooks"]

        if "deps" in hooks and hooks["deps"] is not None:
            for dep_type, dep_object in hooks["deps"].items():
                if dep_object and not isinstance(dep_object, dict):
                    encoded_metadata["hooks"]["deps"][dep_type] = dep_object.to_dict()

        # call_before/after
        if "call_before" in hooks and hooks["call_before"] is not None:
            for i, dep in enumerate(hooks["call_before"]):
                if not isinstance(dep, dict):
                    encoded_metadata["hooks"]["call_before"][i] = dep.to_dict()

        if "call_after" in hooks and hooks["call_after"] is not None:
            for i, dep in enumerate(hooks["call_after"]):
                if not isinstance(dep, dict):
                    encoded_metadata["hooks"]["call_after"][i] = dep.to_dict()

    # triggers
    if "triggers" in metadata:
        if isinstance(metadata["triggers"], list):
            encoded_metadata["triggers"] = []
            for tr in metadata["triggers"]:
                if isinstance(tr, dict):
                    encoded_metadata["triggers"].append(tr)
                else:
                    encoded_metadata["triggers"].append(tr.to_dict())
        else:
            encoded_metadata["triggers"] = metadata["triggers"]

    # qelectron_data_exists
    encoded_metadata["qelectron_data_exists"] = False

    return encoded_metadata


@contextmanager
def pickle_modules_by_value(metadata):
    """
    Pickle modules in a context manager by value.

    Args:
        call_before: The call before metadata.

    Returns:
        None
    """

    call_before = metadata.get("hooks", {}).get("call_before")

    if not call_before:
        yield metadata
        return

    new_metadata = deepcopy(metadata)
    call_before = new_metadata.get("hooks", {}).get("call_before")

    list_of_modules = []

    for i in range(len(call_before)):
        # Extract the pickled module if the call before is a DepsModule
        if call_before[i]["short_name"] == "depsmodule":
            pickled_module = call_before[i]["attributes"]["pickled_module"]
            list_of_modules.append(
                TransportableObject.from_dict(pickled_module).get_deserialized()
            )

            # Delete the DepsModule from new_metadata of the electron
            new_metadata["hooks"]["call_before"][i] = None

    # Remove the None values from the call_before list of this electron
    new_metadata["hooks"]["call_before"] = list(filter(None, new_metadata["hooks"]["call_before"]))

    for module in list_of_modules:
        cloudpickle.register_pickle_by_value(module)

    yield new_metadata

    for module in list_of_modules:
        try:
            cloudpickle.unregister_pickle_by_value(module)
        except ValueError:
            continue


@contextmanager
def add_module_deps_to_lattice_metadata(pp, bound_electrons: Dict[int, Any]):
    old_lattice_metadata = deepcopy(pp.lattice.metadata)

    # Add the module dependencies to the lattice metadata
    for electron in bound_electrons.values():
        call_before = electron.metadata.get("hooks", {}).get("call_before")
        if call_before:
            for i in range(len(call_before)):
                if call_before[i]["short_name"] == "depsmodule":
                    if "hooks" not in pp.lattice.metadata:
                        pp.lattice.metadata["hooks"] = {}

                    if "call_before" not in pp.lattice.metadata["hooks"]:
                        pp.lattice.metadata["hooks"]["call_before"] = []

                    # Temporarily add the module metadat to the lattice metadata for postprocessing
                    pp.lattice.metadata["hooks"]["call_before"].append(call_before[i])

                    # Delete the DepsModule from the electron metadata
                    electron.metadata["hooks"]["call_before"][i] = None

            # Remove the None values from the call_before list of this electron
            electron.metadata["hooks"]["call_before"] = list(
                filter(None, electron.metadata["hooks"]["call_before"])
            )

    yield

    # Restore the old lattice metadata
    pp.lattice.metadata = old_lattice_metadata


class _TransportGraph:
    """
    A TransportGraph is the most essential part of the whole workflow. This contains
    all the information about each electron and lattice required for determining how,
    when, and where to execute the workflow. The TransportGraph contains a directed graph
    which is used to determine the execution order of the nodes. Each node in this graph
    is an electron which is ready to be executed.

    Attributes:
        _graph: The directed graph object of type networkx.DiGraph().
        lattice_metadata: The lattice metadata of the transport graph.
    """

    def __init__(self) -> None:
        self._graph = nx.MultiDiGraph()
        self.lattice_metadata = None

        # IDs of nodes modified during the workflow run
        self.dirty_nodes = []

        self._default_node_attrs = {
            "start_time": None,
            "end_time": None,
            "status": RESULT_STATUS.NEW_OBJECT,
            "output": TransportableObject(None),
            "error": "",
            "sub_dispatch_id": None,
            "sublattice_result": None,
            "stdout": "",
            "stderr": "",
        }

    def add_node(
        self, name: str, function: Callable, metadata: Dict, task_group_id: int = None, **attr
    ) -> int:
        """
        Adds a node to the graph.
        Also serializes the received function.

        Args:
            name: The name of the node.
            function: The function to be executed.
            metadata: The metadata of the node.
            task_group_id: The task group id of the node.
            attr: Any other attributes that need to be added to the node.

        Returns:
            node_key: The node id.

        """
        node_id = len(self._graph.nodes)

        if task_group_id is None:
            task_group_id = node_id

        with pickle_modules_by_value(metadata) as new_metadata:
            serialized_function = TransportableObject(function)

        # Default to gid=node_id

        self._graph.add_node(
            node_id,
            task_group_id=task_group_id,
            name=name,
            function=serialized_function,
            metadata=new_metadata,  # Save the new metadata without the DepsModules in it
            **attr,
        )

        return node_id

    def add_edge(self, x: int, y: int, edge_name: Any, **attr) -> None:
        """
        Adds an edge to the graph and assigns a name to it. Edge insertion
        order is not preserved in networkx. So in case of positional arguments
        passed into the electron, we need to preserve the order when we
        deserialize the request in the lattice.

        Args:
            x: The node id for first node.
            y: The node id for second node.
            edge_name: The name to be assigned to the edge.
            attr: Any other attributes that need to be added to the edge.

        Returns:
            None

        Raises:
            ValueError: If the edge already exists.
        """

        self._graph.add_edge(x, y, edge_name=edge_name, **attr)

    def reset(self) -> None:
        """
        Resets the graph.

        Args:
            None

        Returns:
            None
        """

        self._graph = nx.MultiDiGraph()

    def get_node_value(self, node_key: int, value_key: str) -> Any:
        """
        Get a specific value from a node depending upon the value key.

        Args:
            node_key: The node id.
            value_key: The value key.

        Returns:
            value: The value from the node stored at the value key.

        Raises:
            KeyError: If the value key or node key is not found.
        """
        return self._graph.nodes[node_key][value_key]

    def set_node_value(self, node_key: str, value_key: int, value: Any) -> None:
        """
        Set a certain value of a node. This allows for saving custom data
        in the graph nodes.

        Args:
            node_key: The node id.
            value_key: The value key.
            value: The value to be set at value_key position of the node.

        Returns:
            None

        Raises:
            KeyError: If the node key is not found.
        """

        self.dirty_nodes.append(node_key)
        self._graph.nodes[node_key][value_key] = value

    def get_edge_data(self, dep_key: int, node_key: int) -> Any:
        """
        Get the metadata for all edges between two nodes.

        Args:
            dep_key: The node id for first node.
            node_key: The node id for second node.

        Returns:
            values: A dict {edge_key : value}

        Raises:
            KeyError: If the edge is not found.
        """

        return self._graph.get_edge_data(dep_key, node_key)

    def get_dependencies(self, node_key: int) -> list:
        """
        Gets the parent node ids of a node.

        Args:
            node_key: The node id.

        Returns:
            parents: The dependencies of the node.
        """

        return list(self._graph.predecessors(node_key))

    def get_internal_graph_copy(self) -> nx.MultiDiGraph:
        """
        Get a copy of the internal directed graph
        to avoid modifying the original graph.

        Args:
            None

        Returns:
            graph: A copy of the internal directed graph.
        """

        return self._graph.copy()

    def reset_node(self, node_id: int) -> None:
        """Reset node values to starting state."""
        node_name = self.get_node_value(node_id, "name")

        for node_attr, default_val in self._default_node_attrs.items():
            # Don't clear precomputed parameter outputs.
            if node_attr == "output" and node_name.startswith(parameter_prefix):
                continue

            self.set_node_value(node_id, node_attr, default_val)

    def serialize(self, metadata_only: bool = False) -> bytes:
        """
        Convert transport graph object to JSON to be used in the workflow scheduler.

        Convert transport graph networkx.DiGraph object into JSON format, filter out
        computation specific attributes and lastly add the lattice metadata. This also
        serializes the function Callable into by base64 encoding the cloudpickled result.

        Args:
            metadata_only: If true, only serialize the metadata.

        Returns:
            str: json string representation of transport graph
        """

        # Convert networkx.DiGraph to a format that can be converted to json .
        data = nx.readwrite.node_link_data(self._graph)

        # process each node
        for idx, node in enumerate(data["nodes"]):
            data["nodes"][idx]["function"] = data["nodes"][idx].pop("function").serialize()

        if metadata_only:
            parameter_node_id = [
                i
                for i, node in enumerate(data["nodes"])
                if node["name"].startswith(parameter_prefix)
            ]

            for node in data["nodes"].copy():
                if node["id"] in parameter_node_id:
                    data["nodes"].remove(node)

            # Remove the non-metadata fields such as 'function', 'name', etc from the scheduler workflow input data.
            for idx, node in enumerate(data["nodes"]):
                for field in data["nodes"][idx].copy():
                    if field != "metadata":
                        data["nodes"][idx].pop(field, None)

            # Remove the non-source-target fields from the scheduler workflow input data.
            for idx, node in enumerate(data["links"]):
                for name in data["links"][idx].copy():
                    if name not in ["source", "target"]:
                        data["links"][idx].pop("edge_name", None)

        data["lattice_metadata"] = self.lattice_metadata
        return cloudpickle.dumps(data)

    def serialize_to_json(self, metadata_only: bool = False) -> str:
        """
        Convert transport graph object to JSON to be used in the workflow scheduler.

        Convert transport graph networkx.DiGraph object into JSON format, filter out
        computation specific attributes and lastly add the lattice metadata. This also
        serializes the function Callable into by base64 encoding the cloudpickled result.

        Args:
            metadata_only: If true, only serialize the metadata.

        Returns:
            str: json string representation of transport graph

        Note: serialize_to_json converts metadata objects into dictionary representations.
        """

        # Convert networkx.DiGraph to a format that can be converted to json .
        data = nx.readwrite.node_link_data(self._graph)

        # process each node
        for idx, node in enumerate(data["nodes"]):
            data["nodes"][idx]["function"] = data["nodes"][idx].pop("function").to_dict()
            if "value" in node:
                node["value"] = node["value"].to_dict()
            if "output" in node:
                node["output"] = node["output"].to_dict()
            if "metadata" in node:
                node["metadata"] = encode_metadata(node["metadata"])
            if "start_time" in node:
                if node["start_time"]:
                    node["start_time"] = node["start_time"].isoformat()
            if "end_time" in node:
                if node["end_time"]:
                    node["end_time"] = node["end_time"].isoformat()
            if "status" in node:
                node["status"] = str(node["status"])

        if metadata_only:
            parameter_node_id = [
                i
                for i, node in enumerate(data["nodes"])
                if node["name"].startswith(parameter_prefix)
            ]

            for node in data["nodes"].copy():
                if node["id"] in parameter_node_id:
                    data["nodes"].remove(node)

            # Remove the non-metadata fields such as 'function', 'name', etc from the scheduler workflow input data.
            for idx, node in enumerate(data["nodes"]):
                for field in data["nodes"][idx].copy():
                    if field != "metadata":
                        data["nodes"][idx].pop(field, None)

            # Remove the non-source-target fields from the scheduler workflow input data.
            for idx, node in enumerate(data["links"]):
                for name in data["links"][idx].copy():
                    if name not in ["source", "target"]:
                        data["links"][idx].pop("edge_name", None)

        data["lattice_metadata"] = encode_metadata(self.lattice_metadata)
        return json.dumps(data)

    def deserialize(self, pickled_data: bytes) -> None:
        """
        Load pickled representation of transport graph into the transport graph instance.

        This overwrites anything currently set in the transport graph and deserializes
        the base64 encoded cloudpickled function into a Callable.

        Args:
            pickled_data: Cloudpickled representation of the transport graph

        Returns:
            None
        """

        node_link_data = cloudpickle.loads(pickled_data)
        if "lattice_metadata" in node_link_data:
            self.lattice_metadata = node_link_data["lattice_metadata"]

        for idx, _ in enumerate(node_link_data["nodes"]):
            function_ser = node_link_data["nodes"][idx].pop("function")
            node_link_data["nodes"][idx]["function"] = TransportableObject.deserialize(
                function_ser
            )
        self._graph = nx.readwrite.node_link_graph(node_link_data)

    def deserialize_from_json(self, json_data: str) -> None:
        """Load JSON representation of transport graph into the transport graph instance.

        This overwrites anything currently set in the transport
        graph. Note that metadata (node and lattice-level) need to be
        reconstituted from their dictionary representations when
        needed.

        Args:
            json_data: JSON representation of the transport graph

        Returns:
            None

        """

        node_link_data = json.loads(json_data)
        if "lattice_metadata" in node_link_data:
            self.lattice_metadata = node_link_data["lattice_metadata"]

        for idx, node in enumerate(node_link_data["nodes"]):
            function_ser = node_link_data["nodes"][idx].pop("function")
            node_link_data["nodes"][idx]["function"] = TransportableObject.from_dict(function_ser)
            if "value" in node:
                node["value"] = TransportableObject.from_dict(node["value"])
            if "output" in node:
                node["output"] = TransportableObject.from_dict(node["output"])
            if "start_time" in node:
                if node["start_time"]:
                    node["start_time"] = datetime.datetime.fromisoformat(node["start_time"])
            if "end_time" in node:
                if node["end_time"]:
                    node["end_time"] = datetime.datetime.fromisoformat(node["end_time"])
            if "status" in node:
                node["status"] = Status(node["status"])

        self._graph = nx.readwrite.node_link_graph(node_link_data)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simple Key-Value store base
"""


class _KeyValueBase:
    async def get(self, key):
        raise NotImplementedError

    async def insert(self, key, val):
        raise NotImplementedError

    async def belongs(self, key):
        raise NotImplementedError

    async def remove(self, key):
        raise NotImplementedError

    async def increment(self, key: str, delta: int) -> int:
        """Increments value for `key` by amount `delta`

        Parameters:
            key: the value to change
            delta: the amount to change (can be negative)
        Returns:
            The new value
        """

        raise NotImplementedError


class _DictStore(_KeyValueBase):
    def __init__(self):
        self._store = {}

    async def get(self, key):
        return self._store[key]

    async def insert(self, key, val):
        self._store[key] = val

    async def belongs(self, key):
        return key in self._store

    async def remove(self, key):
        del self._store[key]

    async def increment(self, key, delta: int):
        self._store[key] += delta
        return self._store[key]



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FastAPI models for /api/v1/resultv2 endpoints"""

from datetime import datetime
from typing import Dict, Optional

from pydantic import BaseModel, field_validator

from .asset import AssetSchema
from .common import StatusEnum

ELECTRON_METADATA_KEYS = {
    "task_group_id",
    "name",
    "start_time",
    "end_time",
    "status",
    # user dependent metadata
    "executor",
    "executor_data",
    "qelectron_data_exists",
}

ELECTRON_ASSET_KEYS = {
    "function",
    "function_string",
    "output",
    "value",
    "error",
    "stdout",
    "stderr",
    "qelectron_db",
    # user dependent assets
    "hooks",
}

ELECTRON_FUNCTION_FILENAME = "function.tobj"
ELECTRON_FUNCTION_STRING_FILENAME = "function_string.txt"
ELECTRON_VALUE_FILENAME = "value.tobj"
ELECTRON_STDOUT_FILENAME = "stdout.log"
ELECTRON_STDERR_FILENAME = "stderr.log"
ELECTRON_QELECTRON_DB_FILENAME = (
    "data.mdb"  # This NEEDS to be data.mdb so that the qelectron's Database class can find it
)
ELECTRON_ERROR_FILENAME = "error.log"
ELECTRON_RESULTS_FILENAME = "results.tobj"
ELECTRON_HOOKS_FILENAME = "hooks.json"
ELECTRON_CALL_BEFORE_FILENAME = "call_before.json"
ELECTRON_CALL_AFTER_FILENAME = "call_after.json"
ELECTRON_STORAGE_TYPE = "file"


ASSET_FILENAME_MAP = {
    "function": ELECTRON_FUNCTION_FILENAME,
    "function_string": ELECTRON_FUNCTION_STRING_FILENAME,
    "value": ELECTRON_VALUE_FILENAME,
    "output": ELECTRON_RESULTS_FILENAME,
    "stdout": ELECTRON_STDOUT_FILENAME,
    "stderr": ELECTRON_STDERR_FILENAME,
    "qelectron_db": ELECTRON_QELECTRON_DB_FILENAME,
    "error": ELECTRON_ERROR_FILENAME,
    "hooks": ELECTRON_HOOKS_FILENAME,
}


class ElectronAssets(BaseModel):
    function: AssetSchema
    function_string: AssetSchema
    value: AssetSchema
    output: AssetSchema
    error: Optional[AssetSchema] = None
    stdout: Optional[AssetSchema] = None
    stderr: Optional[AssetSchema] = None
    qelectron_db: Optional[AssetSchema] = None

    # user dependent assets
    hooks: AssetSchema


class ElectronMetadata(BaseModel):
    task_group_id: int
    name: str
    executor: str
    executor_data: dict
    qelectron_data_exists: Optional[bool] = None
    sub_dispatch_id: Optional[str] = None
    status: StatusEnum
    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None

    # For use by redispatch
    def reset(self):
        self.status = StatusEnum.NEW_OBJECT
        self.start_time = None
        self.end_time = None


class ElectronSchema(BaseModel):
    id: int
    metadata: ElectronMetadata
    assets: ElectronAssets
    custom_assets: Optional[Dict[str, AssetSchema]] = None

    @field_validator("custom_assets")
    def check_custom_asset_keys(cls, v):
        if v is not None:
            for key in v:
                if key in ASSET_FILENAME_MAP:
                    raise ValueError(f"Asset {key} conflicts with built-in key")
        return v



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import pennylane as qml

from .local_client import LocalQClient


class MiddleWare:
    def __init__(self) -> None:
        self.qclient = LocalQClient()

    def __new__(cls):
        # Making this a singleton class
        if not hasattr(cls, "instance"):
            cls.instance = super(MiddleWare, cls).__new__(cls)
        return cls.instance

    # The following attributes are properties
    # because the qclient might change over time
    # and every time it gets changed, we shouldn't
    # have to set things like:
    # self.database = self.qclient.database
    # Thus, we override the access of these attributes
    # and return/set these dynamically depending upon
    # what the qclient is at that point in time.

    @property
    def selector(self):
        return self.qclient.selector

    @selector.setter
    def selector(self, selector_func):
        self.qclient.selector = selector_func

    @property
    def database(self):
        return self.qclient.database

    def run_circuits_async(
        self,
        qscripts: List[qml.tape.qscript.QuantumScript],
        executors,
        qelectron_info,
        qnode_specs,
    ):
        return self.qclient.submit(qscripts, executors, qelectron_info, qnode_specs)

    def get_results(self, batch_id):
        return self.qclient.get_results(batch_id)


middleware = MiddleWare()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from datetime import datetime, timezone
from pathlib import Path

from sqlalchemy.orm import Session

from covalent._results_manager import Result
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.schemas import electron, lattice, result
from covalent._workflow.transportable_object import TransportableObject

from .._dal.asset import Asset
from .._dal.electron import Electron
from .._dal.job import Job
from .._dal.lattice import Lattice
from .._object_store.local import local_store
from . import models
from .datastore import workflow_db
from .write_result_to_db import get_electron_type, transaction_upsert_electron_dependency_data

app_log = logger.app_log

ELECTRON_FILENAMES = electron.ASSET_FILENAME_MAP
LATTICE_FILENAMES = lattice.ASSET_FILENAME_MAP.copy()
LATTICE_FILENAMES.update(result.ASSET_FILENAME_MAP.copy())

ELECTRON_FUNCTION_FILENAME = ELECTRON_FILENAMES["function"]
ELECTRON_FUNCTION_STRING_FILENAME = ELECTRON_FILENAMES["function_string"]
ELECTRON_VALUE_FILENAME = ELECTRON_FILENAMES["value"]
ELECTRON_STDOUT_FILENAME = ELECTRON_FILENAMES["stdout"]
ELECTRON_STDERR_FILENAME = ELECTRON_FILENAMES["stderr"]
ELECTRON_QELECTRON_DB_FILENAME = ELECTRON_FILENAMES["qelectron_db"]
ELECTRON_ERROR_FILENAME = ELECTRON_FILENAMES["error"]
ELECTRON_RESULTS_FILENAME = ELECTRON_FILENAMES["output"]
ELECTRON_HOOKS_FILENAME = ELECTRON_FILENAMES["hooks"]
ELECTRON_STORAGE_TYPE = "file"
LATTICE_FUNCTION_FILENAME = LATTICE_FILENAMES["workflow_function"]
LATTICE_FUNCTION_STRING_FILENAME = LATTICE_FILENAMES["workflow_function_string"]
LATTICE_DOCSTRING_FILENAME = LATTICE_FILENAMES["doc"]
LATTICE_ERROR_FILENAME = LATTICE_FILENAMES["error"]
LATTICE_INPUTS_FILENAME = LATTICE_FILENAMES["inputs"]
LATTICE_NAMED_ARGS_FILENAME = LATTICE_FILENAMES["named_args"]
LATTICE_NAMED_KWARGS_FILENAME = LATTICE_FILENAMES["named_kwargs"]
LATTICE_RESULTS_FILENAME = LATTICE_FILENAMES["result"]
LATTICE_HOOKS_FILENAME = LATTICE_FILENAMES["hooks"]
LATTICE_COVA_IMPORTS_FILENAME = LATTICE_FILENAMES["cova_imports"]
LATTICE_LATTICE_IMPORTS_FILENAME = LATTICE_FILENAMES["lattice_imports"]
LATTICE_STORAGE_TYPE = "file"

CUSTOM_ASSETS_FIELD = "custom_asset_keys"


def _lattice_data(session: Session, result: Result, electron_id: int = None) -> int:
    """
    Private method to update lattice data in database

    Arg(s)
        session: SQLalchemy session object
        result: Result object associated with the lattice
        electron_id: electron id in the lattice

    Return(s)
        None
    """

    try:
        workflow_func_string = result.lattice.workflow_function_string
    except AttributeError:
        workflow_func_string = None

    # Store all lattice info that belongs in filenames in the results directory
    results_dir = os.environ.get("COVALENT_DATA_DIR") or get_config("dispatcher.results_dir")
    data_storage_path = os.path.join(results_dir, result.dispatch_id)

    assets = {}

    # Ensure that a dispatch is only persisted once
    lattice_recs = Lattice.meta_type.get(
        session,
        fields={"id", "dispatch_id"},
        equality_filters={"dispatch_id": result.dispatch_id},
        membership_filters={},
    )
    if lattice_recs:
        raise RuntimeError("Dispatch already exists in the DB")

    for key, filename, data in [
        ("workflow_function", LATTICE_FUNCTION_FILENAME, result.lattice.workflow_function),
        ("workflow_function_string", LATTICE_FUNCTION_STRING_FILENAME, workflow_func_string),
        ("doc", LATTICE_DOCSTRING_FILENAME, result.lattice.__doc__),
        ("error", LATTICE_ERROR_FILENAME, result.error),
        ("inputs", LATTICE_INPUTS_FILENAME, result.lattice.inputs),
        ("named_args", LATTICE_NAMED_ARGS_FILENAME, result.lattice.named_args),
        ("named_kwargs", LATTICE_NAMED_KWARGS_FILENAME, result.lattice.named_kwargs),
        ("result", LATTICE_RESULTS_FILENAME, result._result),
        ("hooks", LATTICE_HOOKS_FILENAME, result.lattice.metadata["hooks"]),
        ("cova_imports", LATTICE_COVA_IMPORTS_FILENAME, result.lattice.cova_imports),
        ("lattice_imports", LATTICE_LATTICE_IMPORTS_FILENAME, result.lattice.lattice_imports),
    ]:
        digest, size = local_store.store_file(data_storage_path, filename, data)
        asset_record_kwargs = {
            "storage_type": LATTICE_STORAGE_TYPE,
            "storage_path": str(data_storage_path),
            "object_key": filename,
            "digest_alg": digest.algorithm,
            "digest": digest.hexdigest,
            "size": size,
        }

        assets[key] = Asset.create(session, insert_kwargs=asset_record_kwargs, flush=True)

    # Get custom asset declarations
    lat_metadata = result.lattice.metadata
    if CUSTOM_ASSETS_FIELD in lat_metadata:
        for key in lat_metadata[CUSTOM_ASSETS_FIELD]:
            asset_record_kwargs = {
                "storage_type": LATTICE_STORAGE_TYPE,
                "storage_path": str(data_storage_path),
                "object_key": f"{key}.data",
                "digest_alg": "",
                "digest": "",
            }
            assets[key] = Asset.create(session, insert_kwargs=asset_record_kwargs, flush=True)

    # Write lattice records to Database
    session.flush()

    lattice_record_kwarg = {
        "dispatch_id": result.dispatch_id,
        "electron_id": electron_id,
        "status": str(result.status),
        "name": result.lattice.__name__,
        "docstring_filename": LATTICE_DOCSTRING_FILENAME,
        "electron_num": result._num_nodes,
        "completed_electron_num": 0,  # None of the nodes have been executed or completed yet.
        "storage_path": str(data_storage_path),
        "storage_type": LATTICE_STORAGE_TYPE,
        "function_filename": LATTICE_FUNCTION_FILENAME,
        "function_string_filename": LATTICE_FUNCTION_STRING_FILENAME,
        "executor": result.lattice.metadata["executor"],
        "executor_data": json.dumps(result.lattice.metadata["executor_data"]),
        "workflow_executor": result.lattice.metadata["workflow_executor"],
        "workflow_executor_data": json.dumps(result.lattice.metadata["workflow_executor_data"]),
        "error_filename": LATTICE_ERROR_FILENAME,
        "inputs_filename": LATTICE_INPUTS_FILENAME,
        "named_args_filename": LATTICE_NAMED_ARGS_FILENAME,
        "named_kwargs_filename": LATTICE_NAMED_KWARGS_FILENAME,
        "results_filename": LATTICE_RESULTS_FILENAME,
        "hooks_filename": LATTICE_HOOKS_FILENAME,
        "cova_imports_filename": LATTICE_COVA_IMPORTS_FILENAME,
        "lattice_imports_filename": LATTICE_LATTICE_IMPORTS_FILENAME,
        "results_dir": results_dir,
        "root_dispatch_id": result.root_dispatch_id,
        "python_version": result.lattice.python_version,
        "covalent_version": result.lattice.covalent_version,
        "created_at": datetime.now(timezone.utc),
        "updated_at": datetime.now(timezone.utc),
        "started_at": result.start_time,
        "completed_at": result.end_time,
    }
    lattice_row = Lattice.meta_type.create(session, insert_kwargs=lattice_record_kwarg, flush=True)
    lattice_record = Lattice(session, lattice_row, bare=True, keys={"id"}, electron_keys={"id"})

    lattice_asset_links = [
        lattice_record.associate_asset(session, key, asset.id) for key, asset in assets.items()
    ]
    session.flush()

    return lattice_row.id


def _electron_data(
    session: Session, lattice_id: int, result: Result, cancel_requested: bool = False
) -> dict:
    """
    Update electron data in database

    Arg(s)
        session: SQLalchemy session object
        result: Result object associated with the lattice
        cancel_requested: Boolean indicating whether electron was requested to be cancelled

    Return(s)
        None
    """

    node_id_eid_map = {}
    tg = result.lattice.transport_graph
    dirty_nodes = set(tg.dirty_nodes)
    tg.dirty_nodes.clear()  # Ensure that dirty nodes list is reset once the data is updated

    # Collect task groups and create a job record for each group
    task_groups = {}
    for node_id in dirty_nodes:
        gid = tg.get_node_value(node_id, "task_group_id")
        if gid not in task_groups:
            task_groups[gid] = [node_id]
        else:
            task_groups[gid].append(node_id)

    timestamp = datetime.now(timezone.utc)

    for gid, nodes in task_groups.items():
        job_row = Job.create(
            session, insert_kwargs={"cancel_requested": cancel_requested}, flush=True
        )

        app_log.debug(f"Created job record for task group {result.dispatch_id}:{gid}")

        for node_id in nodes:
            results_dir = os.environ.get("COVALENT_DATA_DIR") or get_config(
                "dispatcher.results_dir"
            )
            node_path = Path(os.path.join(results_dir, result.dispatch_id, f"node_{node_id}"))

            if not node_path.exists():
                node_path.mkdir()

            node_name = tg.get_node_value(node_id, "name")

            try:
                function_string = tg.get_node_value(node_id, "function_string")
            except KeyError:
                function_string = None

            try:
                node_value = tg.get_node_value(node_id, "value")
            except KeyError:
                node_value = TransportableObject(None)

            try:
                node_stdout = tg.get_node_value(node_id, "stdout")
            except KeyError:
                node_stdout = None

            try:
                node_stderr = tg.get_node_value(node_id, "stderr")
            except KeyError:
                node_stderr = None

            try:
                node_error = tg.get_node_value(node_id, "error")
            except KeyError:
                node_error = None

            try:
                node_output = tg.get_node_value(node_id, "output")
            except KeyError:
                node_output = TransportableObject(None)

            try:
                node_qelectron_data_exists = tg.get_node_value(node_id, "qelectron_data_exists")
            except KeyError:
                node_qelectron_data_exists = False

            try:
                node_qelectron_data = tg.get_node_value(node_id, "qelectron_db")
            except KeyError:
                node_qelectron_data = bytes()

            executor = tg.get_node_value(node_id, "metadata")["executor"]
            started_at = tg.get_node_value(node_key=node_id, value_key="start_time")
            completed_at = tg.get_node_value(node_key=node_id, value_key="end_time")

            assets = {}

            for key, filename, data in [
                ("function", ELECTRON_FUNCTION_FILENAME, tg.get_node_value(node_id, "function")),
                ("function_string", ELECTRON_FUNCTION_STRING_FILENAME, function_string),
                ("value", ELECTRON_VALUE_FILENAME, node_value),
                (
                    "hooks",
                    ELECTRON_HOOKS_FILENAME,
                    tg.get_node_value(node_id, "metadata")["hooks"],
                ),
                ("stdout", ELECTRON_STDOUT_FILENAME, node_stdout),
                ("stderr", ELECTRON_STDERR_FILENAME, node_stderr),
                ("qelectron_db", ELECTRON_QELECTRON_DB_FILENAME, node_qelectron_data),
                ("error", ELECTRON_ERROR_FILENAME, node_error),
                ("output", ELECTRON_RESULTS_FILENAME, node_output),
            ]:
                digest, size = local_store.store_file(node_path, filename, data)
                asset_record_kwargs = {
                    "storage_type": ELECTRON_STORAGE_TYPE,
                    "storage_path": str(node_path),
                    "object_key": filename,
                    "digest_alg": digest.algorithm,
                    "digest": digest.hexdigest,
                    "size": size,
                }

                assets[key] = Asset.create(session, insert_kwargs=asset_record_kwargs, flush=True)

            # Register custom assets
            node_metadata = tg.get_node_value(node_id, "metadata")
            if CUSTOM_ASSETS_FIELD in node_metadata:
                for key in node_metadata[CUSTOM_ASSETS_FIELD]:
                    asset_record_kwargs = {
                        "storage_type": LATTICE_STORAGE_TYPE,
                        "storage_path": str(node_path),
                        "object_key": f"{key}.data",
                        "digest_alg": "",
                        "digest": "",
                    }
                    assets[key] = Asset.create(
                        session, insert_kwargs=asset_record_kwargs, flush=True
                    )

            status = tg.get_node_value(node_key=node_id, value_key="status")
            executor_data = tg.get_node_value(node_id, "metadata")["executor_data"]

            electron_record_kwarg = {
                "parent_lattice_id": lattice_id,
                "transport_graph_node_id": node_id,
                "task_group_id": gid,
                "type": get_electron_type(tg.get_node_value(node_key=node_id, value_key="name")),
                "name": node_name,
                "status": str(status),
                "storage_type": ELECTRON_STORAGE_TYPE,
                "storage_path": str(node_path),
                "function_filename": ELECTRON_FUNCTION_FILENAME,
                "function_string_filename": ELECTRON_FUNCTION_STRING_FILENAME,
                "executor": executor,
                "executor_data": json.dumps(executor_data),
                "results_filename": ELECTRON_RESULTS_FILENAME,
                "value_filename": ELECTRON_VALUE_FILENAME,
                "stdout_filename": ELECTRON_STDOUT_FILENAME,
                "stderr_filename": ELECTRON_STDERR_FILENAME,
                "error_filename": ELECTRON_ERROR_FILENAME,
                "hooks_filename": ELECTRON_HOOKS_FILENAME,
                "qelectron_data_exists": node_qelectron_data_exists,
                "job_id": job_row.id,
                "created_at": timestamp,
                "updated_at": timestamp,
                "started_at": started_at,
                "completed_at": completed_at,
            }
            electron_row = Electron.meta_type.create(
                session,
                insert_kwargs=electron_record_kwarg,
                flush=True,
            )
            electron_record = Electron(session, electron_row, keys={"id"})

            node_id_eid_map[node_id] = electron_row.id

            electron_asset_links = [
                electron_record.associate_asset(session, key, asset.id)
                for key, asset in assets.items()
            ]
            session.flush()

    return node_id_eid_map


def persist_result(result: Result, electron_id: int = None) -> None:
    """
    Persist the result object of the lattice recursively into the database

    Arg(s)
        result: Result object associated with the lattice
        electron_id: ID of the electron within the lattice

    Return(s)
        None
    """
    with workflow_db.session() as session:
        parent_lattice_id = _lattice_data(session, result, electron_id)
        if electron_id:
            e_record = (
                session.query(models.Electron).where(models.Electron.id == electron_id).first()
            )
            job_record = Job.get_by_primary_key(session, e_record.job_id)
            cancel_requested = job_record.cancel_requested
        else:
            cancel_requested = False
        node_id_eid_map = _electron_data(session, parent_lattice_id, result, cancel_requested)
        transaction_upsert_electron_dependency_data(session, result.dispatch_id, result.lattice)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from os import environ, path
from pathlib import Path
from typing import Generator, Optional

from alembic import command
from alembic.config import Config
from alembic.environment import EnvironmentContext
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy_utils import create_database, database_exists

from covalent._shared_files.config import get_config

from . import models

DEBUG_DB = environ.get("COVALENT_DEBUG_DB") == "1"


class DataStore:
    def __init__(
        self,
        db_URL: Optional[str] = None,
        initialize_db: bool = False,
        **kwargs,
    ):
        if db_URL:
            self.db_URL = db_URL
        else:
            self.db_URL = "sqlite+pysqlite:///" + get_config("dispatcher.db_path")

        self.engine = create_engine(self.db_URL, **kwargs)
        if not database_exists(self.engine.url):
            try:
                create_database(self.engine.url)
            except Exception:
                pass
        self.Session = sessionmaker(self.engine)

        # flag should only be used in pytest - tables should be generated using migrations
        if initialize_db:
            models.Base.metadata.create_all(self.engine)

    @staticmethod
    def factory():
        return DataStore(db_URL=environ.get("COVALENT_DATABASE_URL"), echo=DEBUG_DB)

    def get_alembic_config(self, logging_enabled: bool = True):
        alembic_ini_path = Path(path.join(__file__, "./../../../covalent_migrations/alembic.ini"))
        migrations_folder = alembic_ini_path / Path("..")
        alembic_config = Config(str(alembic_ini_path.resolve()))
        alembic_config.set_main_option("script_location", str(migrations_folder.resolve()))
        alembic_config.attributes["configure_logger"] = logging_enabled
        alembic_config.set_main_option("sqlalchemy.url", self.db_URL)
        return alembic_config

    def run_migrations(self, logging_enabled: bool = True):
        alembic_config = self.get_alembic_config(logging_enabled=logging_enabled)
        command.upgrade(alembic_config, "head")

    def current_revision(self):
        alembic_config = self.get_alembic_config(logging_enabled=False)
        script = ScriptDirectory.from_config(alembic_config)
        with EnvironmentContext(alembic_config, script) as env_ctx:
            migration_ctx = MigrationContext.configure(
                self.engine.connect(), environment_context=env_ctx
            )
            current_rev = migration_ctx.get_current_revision()
            return current_rev

    def current_head(self):
        alembic_config = self.get_alembic_config(logging_enabled=False)
        script = ScriptDirectory.from_config(alembic_config)
        current_head = script.get_current_head()
        return current_head

    @property
    def is_migration_pending(self):
        return self.current_head() != self.current_revision()

    @contextmanager
    def session(self) -> Generator[Session, None, None]:
        with self.Session.begin() as session:
            yield session


workflow_db = DataStore.factory()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from furl import furl

from ..._shared_files import logger
from .. import File
from .transfer_strategy_base import FileTransferStrategy

app_log = logger.app_log
log_stack_info = logger.log_stack_info


class S3(FileTransferStrategy):

    """
    Implements Base FileTransferStrategy class to upload/download files from S3 Bucket.
    """

    def __init__(self, credentials: str = None, profile: str = None, region_name: str = None):
        self.credentials = credentials
        self.profile = profile
        self.region_name = region_name

        try:
            import boto3
        except ImportError:
            raise ImportError(
                "Using S3 strategy requires boto3 from AWS installed on your system."
            )

        if self.credentials:
            os.environ["AWS_SHARED_CREDENTIALS_FILE"] = self.credentials

    # return callable to download here implies 'from' is a remote source
    def download(self, from_file: File, to_file: File = File()) -> File:
        """Download files or the contents of folders from S3 bucket."""
        app_log.debug(f"Is dir: {from_file._is_dir}")

        from_filepath = from_file.filepath
        to_filepath = to_file.filepath

        if from_filepath.startswith("/"):
            from_filepath = from_filepath[1:]

        bucket_name = furl(from_file.uri).origin[5:]
        app_log.debug(
            f"S3 download bucket: {bucket_name}, from_filepath: {from_filepath}, to_filepath {to_filepath}."
        )

        executor_profile = self.profile
        executor_region = self.region_name

        def get_boto_options(profile=None, region=None):
            boto_options = {}
            if profile:
                boto_options["profile_name"] = profile
            if region:
                boto_options["region_name"] = region
            return boto_options

        if from_file._is_dir:

            def callable():
                """Download files from a folder in s3 bucket."""
                from pathlib import Path

                import boto3

                profile = executor_profile
                region = executor_region
                s3 = boto3.Session(**get_boto_options(profile, region)).client("s3")

                for obj_metadata in s3.list_objects(Bucket=bucket_name, Prefix=from_filepath)[
                    "Contents"
                ]:
                    obj_key = obj_metadata["Key"]
                    obj_destination_filepath = Path(to_filepath) / obj_key
                    if not obj_key.endswith("/"):
                        obj_destination_filepath.parents[0].mkdir(parents=True, exist_ok=True)
                        app_log.debug(
                            f"Downloading file {str(Path(from_filepath) / obj_key)} to {str(obj_destination_filepath)}."
                        )
                        s3.download_file(
                            bucket_name,
                            str(Path(from_filepath) / obj_key),
                            str(obj_destination_filepath),
                        )

        else:

            def callable():
                """Download file from s3 bucket."""
                import boto3

                profile = executor_profile
                region = executor_region
                s3 = boto3.Session(**get_boto_options(profile, region)).client("s3")

                s3.download_file(bucket_name, from_filepath, to_filepath)

        return callable

    # return callable to download here implies 'to' is a remote source
    def upload(self, from_file: File, to_file: File = File()) -> File:
        """Upload files or folders to S3 bucket."""
        from_filepath = from_file.filepath
        to_filepath = to_file.filepath

        if to_filepath.startswith("/"):
            to_filepath = to_filepath.strip("/")

        bucket_name = furl(to_file.uri).origin[5:]
        app_log.debug(
            f"S3 upload bucket: {bucket_name}, from_filepath: {from_filepath}, to_filepath {to_filepath}."
        )

        executor_profile = self.profile
        executor_region = self.region_name

        def get_boto_options(profile=None, region=None):
            boto_options = {}
            if profile:
                boto_options["profile_name"] = profile
            if region:
                boto_options["region_name"] = region
            return boto_options

        if from_file._is_dir:

            def callable():
                """List and upload files from a directory to the remote S3 bucket."""
                import os
                from pathlib import Path

                import boto3

                profile = executor_profile
                region = executor_region
                s3 = boto3.Session(**get_boto_options(profile, region)).client("s3")

                for dir_, _, files in os.walk(from_filepath):
                    for file_name in files:
                        rel_dir = os.path.relpath(dir_, from_filepath)
                        rel_file = os.path.join(rel_dir, file_name)
                        obj_from_filepath = str(Path(from_filepath) / rel_file)
                        obj_to_filepath = str(Path(to_filepath) / rel_file)
                        app_log.debug(f"Uploading from {obj_from_filepath} to {obj_to_filepath}.")
                        s3.upload_file(obj_from_filepath, bucket_name, obj_to_filepath)

        else:

            def callable():
                """Upload file to remote S3 bucket."""
                import boto3

                profile = executor_profile
                region = executor_region
                s3 = boto3.Session(**get_boto_options(profile, region)).client("s3")

                s3.upload_file(from_filepath, bucket_name, to_filepath)

        return callable

    # No S3 Strategy for copy
    def cp(self, from_file: File, to_file: File = File()) -> File:
        raise NotImplementedError



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from abc import ABC, abstractmethod
from concurrent.futures import Future
from typing import Callable, List, Sequence, Union

from mpire.async_result import AsyncResult
from pydantic import BaseModel, ConfigDict

from ...executor.qbase import AsyncBaseQExecutor, BaseQExecutor, QCResult


class AsyncBaseQCluster(AsyncBaseQExecutor):
    executors: Sequence[BaseQExecutor]
    selector: Union[str, Callable]

    # Flag used to indicate whether `self.selector` is currently serialized.
    # This needs to be without the "_" prefix so that it gets propagated to the server.
    selector_serialized: bool = False

    @abstractmethod
    def serialize_selector(self) -> None:
        """
        Serializes the cluster's selector function.
        """
        raise NotImplementedError

    @abstractmethod
    def deserialize_selector(self) -> Union[str, Callable]:
        """
        Deserializes the cluster's selector function.
        """
        raise NotImplementedError

    @abstractmethod
    def dict(self, *args, **kwargs) -> dict:
        """
        Custom dict method to create a hashable `executors` attribute.
        """
        raise NotImplementedError

    @abstractmethod
    def get_selector(self):
        """
        Returns the deserialized selector function.
        """
        raise NotImplementedError

    async def _get_result(self, futures_list: List) -> List[QCResult]:
        """
        Override the base method to handle the case where the `futures_list`
        contains a mix of object types from various executors.
        """
        results_and_times = []
        for fut in futures_list:
            if isinstance(fut, asyncio.Task):
                results_and_times.append(await fut)
            elif isinstance(fut, Future):
                results_and_times.append(fut.result())
            elif isinstance(fut, AsyncResult):
                results_and_times.append(fut.get())
            else:
                results_and_times.append(fut)

        return results_and_times


class BaseQSelector(ABC, BaseModel):
    name: str = "base_qselector"

    def __call__(self, qscript, executors):
        """ "
        Interface used by the quantum server.
        """
        return self.selector_function(qscript, executors)

    @abstractmethod
    def selector_function(self, qscript, executors):
        """
        Implement selection logic here.
        """
        raise NotImplementedError

    model_config = ConfigDict(extra="allow")



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dispatch request and response model"""


from datetime import datetime
from typing import List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Annotated

from covalent_ui.api.v1.utils.models_helper import SortBy, SortDirection
from covalent_ui.api.v1.utils.status import Status


class DispatchSummaryRequest(BaseModel):
    """Dispatch Summary Request model"""

    count: Annotated[int, Field(gt=0, lt=100)]
    offset: Optional[Annotated[int, Field(gt=-1)]] = 0
    sort_by: Optional[SortBy] = SortBy.STARTED
    search: Optional[str] = ""
    direction: Optional[SortDirection] = SortDirection.DESCENDING
    filter: Optional[Status] = Status.ALL


class DispatchModule(BaseModel):
    """Dispatch Modeule Validation"""

    dispatch_id: str
    lattice_name: str
    runtime: Optional[Union[int, float, None]] = None
    total_electrons: Optional[Union[int, None]] = None
    total_electrons_completed: Optional[Union[int, None]] = None
    started_at: Optional[Union[datetime, None]] = None
    ended_at: Optional[Union[datetime, None]] = None
    status: Status
    updated_at: Optional[Union[datetime, None]] = None
    model_config = ConfigDict(from_attributes=True)


class DispatchResponse(BaseModel):
    """Dispatch Response Model"""

    items: List[DispatchModule]
    total_count: int
    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "dispatches": [
                    {
                        "dispatch_id": "1b44989a-1c65-4148-959e-00062a34ac16",
                        "lattice_name": "testing content",
                        "runtime": 1,
                        "started_time": "2022-06-13T07:45:02.114328+00:00",
                        "end_time": "2022-06-13T07:45:02.216474+00:00",
                        "status": "COMPLETED",
                    }
                ],
                "total_count": 10,
            }
        }
    )


class DeleteDispatchesRequest(BaseModel):
    """Dashboard delete request model"""

    dispatches: Optional[List[UUID]] = None


class DeleteAllDispatchesRequest(BaseModel):
    """Dashboard delete all request model"""

    status_filter: Optional[Status] = Status.ALL
    search_string: Optional[str] = ""


class DeleteDispatchesResponse(BaseModel):
    """Dashboard metadate model"""

    success_items: List[UUID]
    failure_items: Union[List[UUID], None] = None
    message: Union[str, None] = None


class DispatchDashBoardResponse(BaseModel):
    """Dashboard metadate model"""

    total_jobs: Union[int, None] = None
    total_jobs_running: Union[int, None] = None
    total_jobs_completed: Union[int, None] = None
    total_jobs_failed: Union[int, None] = None
    total_jobs_cancelled: Union[int, None] = None
    total_jobs_new_object: Union[int, None] = None
    latest_running_task_status: Union[Status, None] = None
    total_dispatcher_duration: Union[int, None] = None
    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "total_jobs": 5,
                "total_jobs_running": 5,
                "total_jobs_completed": 20,
                "total_jobs_failed": 3,
                "total_jobs_cancelled": 0,
                "total_jobs_new_object": 1,
                "latest_running_task_status": "COMPLETED",
                "total_dispatcher_duration": 90,
            }
        }
    )



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""File handlers"""

import base64
import json

import cloudpickle as pickle

from covalent._workflow.transport import TransportableObject, _TransportGraph
from covalent_dispatcher._dal.asset import local_store


def transportable_object(obj):
    """Decode transportable object
    Args:
        obj: Covalent transportable object
    Returns:
        Decoded transportable object
    """
    if obj:
        load_pickle = base64.b64decode(obj._object.encode("utf-8"))
        return f"\npickle.loads({load_pickle})"
    return None


def validate_data(unpickled_object):
    """Validate unpickled object"""
    if isinstance(unpickled_object, list):
        if not (unpickled_object):
            return ""
        else:
            list_str = "".join([obj for obj in unpickled_object])
            return list_str
    if isinstance(unpickled_object, dict):
        args_array = []
        kwargs_array = {}
        if bool(unpickled_object):
            if "type" in unpickled_object:
                return unpickled_object
            for obj in unpickled_object["args"]:
                args_array.append(obj.object_string) if obj is not None else None

            for obj in unpickled_object["kwargs"]:
                kwargs_array[obj] = (
                    unpickled_object["kwargs"][obj].object_string
                    if unpickled_object["kwargs"][obj] is not None
                    else None
                )

            to_transportable_object = TransportableObject(
                {"args": tuple(args_array), "kwargs": kwargs_array}
            )
            object_bytes = transportable_object(to_transportable_object)
            return (
                str(({"args": tuple(args_array), "kwargs": kwargs_array})),
                f"import pickle{object_bytes}",
            )
        else:
            return None
    elif isinstance(unpickled_object, str):
        return (
            unpickled_object if (unpickled_object != "" or unpickled_object is not None) else None
        )
    elif isinstance(unpickled_object, TransportableObject):
        object_bytes = transportable_object(unpickled_object)
        res = unpickled_object.object_string
        return (
            json.dumps(res),
            f"import pickle{object_bytes}",
        )
    elif isinstance(unpickled_object, _TransportGraph):
        return str(unpickled_object.__dict__)
    else:
        return unpickled_object, unpickled_object


class FileHandler:
    """File read"""

    def __init__(self, location) -> None:
        self.location = location

    def read_from_pickle(self, path):
        """Return data from pickle file"""
        try:
            unpickled_object = self.__unpickle_file(path)
            res = validate_data(unpickled_object)
            return res
        except Exception:
            return None

    def read_from_serialized(self, path):
        """Return data from serialized object"""
        try:
            deserialized_obj = local_store.load_file(self.location, path)
            return validate_data(deserialized_obj)
        except Exception as e:
            return None

    def read_from_text(self, path):
        """Return data from text file"""
        try:
            with open(self.location + "/" + path, "r", encoding="utf-8") as read_file:
                text_object = read_file.read()
                read_file.close()
                return text_object if text_object is not None else ""
        except Exception:
            return None

    def __unpickle_file(self, path):
        try:
            with open(self.location + "/" + path, "rb") as read_file:
                unpickled_object = pickle.load(read_file)
                read_file.close()
                return unpickled_object
        except Exception:
            return None



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Types defining the runner-executor interface
"""

from typing import Dict, List

from pydantic import BaseModel, field_validator

from covalent._shared_files.schemas.asset import AssetUpdate
from covalent._shared_files.util_classes import RESULT_STATUS, Status


class TaskUpdate(BaseModel):
    """Represents a task status update.

    Attributes:
        dispatch_id: The id of the dispatch.
        node_id: The id of the task.
        status: A Status dataclass representing the task's terminal status.
        assets: A map from asset keys to AssetUpdate objects
    """

    dispatch_id: str
    node_id: int
    status: Status
    assets: Dict[str, AssetUpdate]

    @field_validator("status")
    def validate_status(cls, v):
        if RESULT_STATUS.is_terminal(v):
            return v
        else:
            raise ValueError(f"Illegal status update {v}")


class TaskSpec(BaseModel):
    """An abstract description of a runnable task.

    Attributes:
        function_id: The `node_id` of the function.
        args_ids: The `node_id`s of the function's args
        kwargs_ids: The `node_id`s of the function's kwargs {key: node_id}
        hooks_id: An opaque string representing the task's hooks.

    The attribute values can be used in conjunction with a
    `ResourceMap` to locate the actual resources in the compute
    environment.
    """

    function_id: int
    args_ids: List[int]
    kwargs_ids: Dict[str, int]


class ResourceMap(BaseModel):

    """Map resource identifiers to URIs.

    The resources may be loaded in the compute environment from these
    URIs.

    Resource identifiers are the attribute values of TaskSpecs. For
    instance, if ts is a `TaskSpec` and rm is the corresponding
    `ResourceMap`, then
    - the serialized function has URI `rm.functions[ts.function_id]`
    - the serialized args have URIs `rm.inputs[ts.args_ids[i]]`
    - the call_before has URI `rm.deps[ts.call_before_id]`

    Attributes:
        functions: A map from node id to the corresponding URI.
        inputs: A map from node id to the corresponding URI
        deps: A map from deps resource ids to their corresponding URIs.

    """

    # Map node_id to URI
    functions: Dict[int, str]

    # Map node_id to URI
    inputs: Dict[int, str]

    # Includes deps, call_before, call_after
    hooks: Dict[int, str]


class TaskGroup(BaseModel):
    """Description of a group of runnable graph nodes.

    Attributes:
        dispatch_id: The `dispatch_id`.
        task_ids: The graph nodes comprising the task group.
        task_group_id: The task group identifier.
    """

    dispatch_id: str
    task_ids: List[int]
    task_group_id: int



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Graph Data Layer"""
from uuid import UUID

from sqlalchemy import text
from sqlalchemy.orm import Session

from covalent_ui.api.v1.database.schema.electron import Electron
from covalent_ui.api.v1.database.schema.electron_dependency import ElectronDependency
from covalent_ui.api.v1.database.schema.lattices import Lattice


class Graph:
    """Graph data access layer"""

    def __init__(self, db_con: Session) -> None:
        self.db_con = db_con

    def get_nodes(self, parent_lattice_id: int):
        """
        Get nodes from parent_lattice_id
        Args:
            parent_lattice_id: Refers to the parent_lattice_id in electron table
        Return:
            graph data with list of nodes
        """
        sql = text(
            """SELECT
            electrons.id as id,
            electrons.name as name,
            electrons.transport_graph_node_id as node_id,
            electrons.started_at,
            electrons.completed_at,
            electrons.status,
            electrons.type,
            electrons.qelectron_data_exists,
            electrons.executor as executor_label,
            (case when electrons.type = 'sublattice'
            then
            (select lattices.dispatch_id from lattices
            where lattices.electron_id = electrons.id)
            else Null
            END
            ) as sublattice_dispatch_id
            from electrons join lattices on electrons.parent_lattice_id = lattices.id
            where lattices.id = :a
        """
        )
        result = self.db_con.execute(sql, {"a": parent_lattice_id}).fetchall()
        return result

    def get_links(self, parent_lattice_id: int):
        """
        Get links from parent_lattice_id
        When parent_lattice_id passed to get links
            then join electrons and electron_dependency
        Args:
            parent_lattice_id: Refers to the parent_lattice_id id in electrons table
        Return:
            graph data with list of links
        """
        return (
            self.db_con.query(
                ElectronDependency.edge_name,
                ElectronDependency.parameter_type,
                ElectronDependency.electron_id.label("target"),
                ElectronDependency.parent_electron_id.label("source"),
                ElectronDependency.arg_index,
            )
            .join(Electron, Electron.id == ElectronDependency.electron_id)
            .filter(Electron.parent_lattice_id == parent_lattice_id)
            .all()
        )

    def get_graph(self, dispatch_id: UUID):
        """
        Get graph data from parent lattice id
        When dispatch id passed to get graph
            Get list of nodes from Electrons table by passing list of latice id
            Get list of links from Electron dependency table by passing in electron
        Args:
            dispatch_id: Refers to the dispatch id from lattices table
        Return:
            graph data with list of nodes and links
        """
        parent_lattice_id = (
            self.db_con.query(Lattice.id).where(Lattice.dispatch_id == str(dispatch_id)).first()
        )
        if parent_lattice_id is not None:
            parrent_id = parent_lattice_id[0]
            nodes = self.get_nodes(parrent_id)
            links = self.get_links(parrent_id)
            return {"dispatch_id": str(dispatch_id), "nodes": nodes, "links": links}
        return None



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Schema updates for new DAL

Revision ID: 1142d81b29b8
Revises: de0a6c0a3e3d
Create Date: 2023-06-18 09:18:31.450740

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
# pragma: allowlist nextline secret
revision = "1142d81b29b8"
# pragma: allowlist nextline secret
down_revision = "de0a6c0a3e3d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "assets",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("storage_type", sa.Text(), nullable=False),
        sa.Column("storage_path", sa.Text(), nullable=False),
        sa.Column("object_key", sa.Text(), nullable=False),
        sa.Column("digest_alg", sa.Text(), nullable=True),
        sa.Column("digest", sa.Text(), nullable=True),
        sa.Column("remote_uri", sa.Text(), nullable=True),
        sa.Column("size", sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "lattice_assets",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("meta_id", sa.Integer(), nullable=False),
        sa.Column("asset_id", sa.Integer(), nullable=False),
        sa.Column("key", sa.String(length=24), nullable=False),
        sa.Index("lattice_assets_idx", "meta_id", "key"),
        sa.ForeignKeyConstraint(["asset_id"], ["assets.id"], name="asset_link"),
        sa.ForeignKeyConstraint(["meta_id"], ["lattices.id"], name="lattice_link"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "electron_assets",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("meta_id", sa.Integer(), nullable=False),
        sa.Column("asset_id", sa.Integer(), nullable=False),
        sa.Column("key", sa.String(length=24), nullable=False),
        sa.Index("electron_assets_idx", "meta_id", "key"),
        sa.ForeignKeyConstraint(["asset_id"], ["assets.id"], name="asset_link"),
        sa.ForeignKeyConstraint(["meta_id"], ["electrons.id"], name="electron_link"),
        sa.PrimaryKeyConstraint("id"),
    )
    with op.batch_alter_table("electron_dependency", schema=None) as batch_op:
        batch_op.create_index("cnode_idx", ["electron_id"], unique=False)
        batch_op.create_index("pnode_idx", ["parent_electron_id"], unique=False)
        batch_op.create_foreign_key(
            "parent_electron_link", "electrons", ["parent_electron_id"], ["id"]
        )

    with op.batch_alter_table("electrons", schema=None) as batch_op:
        batch_op.add_column(sa.Column("task_group_id", sa.Integer(), nullable=False))
        batch_op.add_column(sa.Column("executor_data", sa.Text(), nullable=True))
        batch_op.create_index(
            "latid_nodeid_idx", ["parent_lattice_id", "transport_graph_node_id"], unique=False
        )
        batch_op.drop_column("executor_data_filename")

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.add_column(sa.Column("status", sa.String(length=24), nullable=False))
        batch_op.drop_column("cancel_successful")

    with op.batch_alter_table("lattices", schema=None) as batch_op:
        batch_op.add_column(sa.Column("executor_data", sa.Text(), nullable=True))
        batch_op.add_column(sa.Column("workflow_executor_data", sa.Text(), nullable=True))
        batch_op.add_column(sa.Column("python_version", sa.Text(), nullable=True))
        batch_op.add_column(sa.Column("covalent_version", sa.Text(), nullable=True))
        batch_op.create_unique_constraint("u_dispatch_id", ["dispatch_id"])
        batch_op.drop_column("executor_data_filename")
        batch_op.drop_column("workflow_executor_data_filename")
        batch_op.drop_column("transport_graph_filename")

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("lattices", schema=None) as batch_op:
        batch_op.add_column(sa.Column("transport_graph_filename", sa.TEXT(), nullable=True))
        batch_op.add_column(sa.Column("workflow_executor_data_filename", sa.TEXT(), nullable=True))
        batch_op.add_column(sa.Column("executor_data_filename", sa.TEXT(), nullable=True))
        batch_op.drop_constraint("u_dispatch_id", type_="unique")
        batch_op.drop_column("covalent_version")
        batch_op.drop_column("python_version")
        batch_op.drop_column("workflow_executor_data")
        batch_op.drop_column("executor_data")

    with op.batch_alter_table("jobs", schema=None) as batch_op:
        batch_op.add_column(sa.Column("cancel_successful", sa.BOOLEAN(), nullable=False))
        batch_op.drop_column("status")

    with op.batch_alter_table("electrons", schema=None) as batch_op:
        batch_op.add_column(sa.Column("executor_data_filename", sa.TEXT(), nullable=True))
        batch_op.drop_index("latid_nodeid_idx")
        batch_op.drop_column("executor_data")
        batch_op.drop_column("task_group_id")

    with op.batch_alter_table("electron_dependency", schema=None) as batch_op:
        batch_op.drop_constraint("parent_electron_link", type_="foreignkey")
        batch_op.drop_index("pnode_idx")
        batch_op.drop_index("cnode_idx")

    op.drop_table("electron_assets")
    op.drop_table("lattice_assets")
    op.drop_table("assets")
    # ### end Alembic commands ###



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main Covalent public functionality."""

import contextlib
from importlib import metadata

from . import _file_transfer as fs  # nopycln: import
from . import executor, leptons  # nopycln: import
from ._dispatcher_plugins import local_dispatch as dispatch  # nopycln: import
from ._dispatcher_plugins import local_dispatch_sync as dispatch_sync  # nopycln: import
from ._dispatcher_plugins import local_redispatch as redispatch  # nopycln: import
from ._dispatcher_plugins import stop_triggers  # nopycln: import
from ._file_transfer import strategies as fs_strategies  # nopycln: import
from ._programmatic.commands import (  # nopycln: import
    covalent_start,
    covalent_stop,
    is_covalent_running,
)
from ._results_manager.results_manager import (  # nopycln: import
    cancel,
    get_result,
    get_result_manager,
)
from ._shared_files.config import get_config, reload_config, set_config  # nopycln: import
from ._shared_files.util_classes import RESULT_STATUS as status  # nopycln: import
from ._workflow import (  # nopycln: import
    DepsBash,
    DepsCall,
    DepsModule,
    DepsPip,
    Lepton,
    TransportableObject,
    electron,
    lattice,
)
from ._workflow.electron import wait  # nopycln: import
from .executor.utils import get_context  # nopycln: import

with contextlib.suppress(ImportError):
    from ._workflow.qelectron import qelectron  # nopycln: import
    from .quantum import QCluster  # nopycln: import

__all__ = [s for s in dir() if not s.startswith("_")]

for _s in dir():
    if not _s.startswith("_"):
        _obj = globals()[_s]
        _obj.__module__ = __name__

__version__ = metadata.version("covalent")



# Copyright 2023 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Define the special QNode that replaces Pennylane circuits decorated with `@ct.qelectron`.
"""

from contextlib import contextmanager
from typing import List

import pennylane as qml

from .._results_manager.qresult import QNodeFutureResult
from .._shared_files import logger
from .._shared_files.qelectron_utils import get_original_shots
from .._shared_files.qinfo import QElectronInfo, QNodeSpecs
from .._shared_files.qresult_utils import re_execute
from ..executor.qbase import BaseQExecutor
from .qdevice import QEDevice

app_log = logger.app_log
log_stack_info = logger.log_stack_info

_GRADIENT_ACCESS_MAXES = {
    "parameter-shift": 2,
}


class QNodeQE(qml.QNode):
    """
    A sub-type of Pennylane's QNode that integrates Covalent QElectrons.

    Attributes:
        original_qnode: The original QNode that was wrapped.
        device: The `QEDevice` instance used for circuit execution instead of the
            original QNode's device.
    """

    def __init__(
        self,
        qnode: qml.QNode,
        executors: List[BaseQExecutor],
        qelectron_info: QElectronInfo,
    ):
        """
        Initialize a `QNodeQE` instance.

        Args:
            qnode: The Pennylane QNode replaced by this object.
            executors: A list of executors to use for circuit execution.
            qelectron_info: Settings related to the original QNode.
        """

        self.original_qnode = qnode

        # Private gradient_fn enables overriding the `gradient_fn` attribute.
        self._gradient_fn = None
        self._gradient_fn_overridden = False

        # Create a new device for every QNodeQE instance
        qe_device = QEDevice(
            wires=qnode.device.num_wires,
            shots=get_original_shots(qnode.device),
            executors=executors,
            qelectron_info=qelectron_info,
        )

        super().__init__(
            func=qnode.func,
            device=qe_device,
            interface=qnode.interface,
            diff_method=qnode.diff_method,
            expansion_strategy=qnode.expansion_strategy,
            max_expansion=qnode.max_expansion,
        )

    @contextmanager
    def mark_call_async(self):
        """
        Activates async execution mode for this instance.
        """
        # pylint: disable=protected-access
        self.device._async_run = True
        try:
            yield
        finally:
            self.device._async_run = False
            self.device._batch_id = None

    def run_later(self, *args, **kwargs) -> QNodeFutureResult:
        """
        Calls this QNode asynchronously. This method returns immediately, without
        waiting for the circuit to finish executing.

        Returns:
            future_result: An object that can be queried for the execution result.
        """

        # pylint: disable=protected-access
        with self.mark_call_async():
            self(*args, **kwargs)
            batch_id = self.device._batch_id

        future_result = QNodeFutureResult(
            batch_id,
            interface=self.device.qnode_specs.interface,
            original_qnode=self.original_qnode,
            original_tape=self.tape,
        )

        return future_result(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        Call QElectrons with args and kwargs suitable for the original QNode.
        """

        self.device.qnode_specs = self._specs(*args, **kwargs)

        # This leads to execution on the QEDevice `self.device`.
        retval = super().__call__(*args, **kwargs)
        self._update_num_executions()

        if self.device._async_run:
            # Do nothing during call from `run_later()`.
            return None

        # Skip re-execution for numpy interface.
        interface = self.device.qnode_specs.interface
        if interface in {"numpy", "autograd"}:
            return retval

        # Run through an overloaded circuit/device to loop `retval` into interfaces.
        retval = re_execute(retval, self.original_qnode, self.tape)(interface, *args, **kwargs)

        return retval

    def _update_num_executions(self):
        """
        Increment number of executions on original and custom device.
        """
        self.device._num_executions += 1
        self.original_qnode.device._num_executions += 1

    def _specs(self, *args, **kwargs) -> QNodeSpecs:
        """
        Check args and kwargs to avoid computing gradients on non-trainable parameters.
        Update the interface if it is set to "auto".
        """

        # Some args or some kwargs are trainable. No warning expected.
        if (args or kwargs) and (
            any(qml.math.get_trainable_indices(args))
            or any(qml.math.get_trainable_indices(kwargs.values()))
        ):
            specs = QNodeSpecs(**qml.specs(self)(*args, **kwargs))
        else:
            # No trainable params. Avoid warning.
            with self.override_gradient_fn(None):
                specs = QNodeSpecs(**qml.specs(self)(*args, **kwargs))

            # Replace override value with actual `gradient_fn`.
            self.construct(args, kwargs)
            specs.gradient_fn = self.gradient_fn

        # This will be done inside QNode.__call__() to update `self.interface`.
        # Here, we anticipate that change and update the specs as well.
        specs.interface = qml.math.get_interface(*args, *list(kwargs.values()))

        return specs

    @contextmanager
    def override_gradient_fn(self, fn):
        """
        Override the private `self._gradient_fn` to override the `gradient_fn`
        attribute (property).
        """
        self._gradient_fn, _tmp_gradient_fn = fn, self._gradient_fn
        self._gradient_fn_overridden = True
        try:
            yield
        finally:
            self._gradient_fn = _tmp_gradient_fn
            self._gradient_fn_overridden = False

    @property
    def gradient_fn(self):
        """
        This property replaces the `qml.QNode.gradient_fn` attribute and enables overriding it.
        """
        return self._gradient_fn

    @gradient_fn.setter
    def gradient_fn(self, fn):
        """
        Enforces gradient override.
        """
        if self._gradient_fn_overridden:
            return  # disallow setting `.gradient_fn``
        self._gradient_fn = fn



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import urllib.request

from .. import File
from .transfer_strategy_base import FileTransferStrategy


class HTTP(FileTransferStrategy):
    """
    Implements Base FileTransferStrategy class to use HTTP to download files from public URLs.
    """

    # return callable to download here implies 'from' is a remote source
    def download(self, from_file: File, to_file: File = File()) -> File:
        from_filepath = from_file.uri
        to_filepath = to_file.filepath

        def callable():
            urllib.request.urlretrieve(from_filepath, to_filepath)
            return to_filepath

        return callable

    # HTTP Strategy is read only
    def upload(self, from_file: File, to_file: File = File()) -> File:
        raise NotImplementedError

    # HTTP Strategy is read only
    def cp(self, from_file: File, to_file: File = File()) -> File:
        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DB-backed electron"""

from typing import Any, Dict, List

from sqlalchemy import select
from sqlalchemy.orm import Session

from .._db import models
from .base import DispatchedObject
from .controller import Record
from .db_interfaces.electron_utils import ASSET_KEYS  # nopycln: import
from .db_interfaces.electron_utils import METADATA_KEYS  # nopycln: import
from .db_interfaces.electron_utils import _meta_record_map, get_filters, set_filters

ELECTRON_KEYS = list(_meta_record_map.keys())


class ElectronMeta(Record[models.Electron]):
    model = models.Electron


class ElectronAsset(Record[models.ElectronAsset]):
    model = models.ElectronAsset


class Electron(DispatchedObject[ElectronMeta, ElectronAsset]):
    meta_type = ElectronMeta
    asset_link_type = ElectronAsset

    metadata_keys = ELECTRON_KEYS

    def __init__(self, session: Session, record: models.Electron, *, keys: List = ELECTRON_KEYS):
        self._id = record.id
        self._keys = keys

        fields = set(map(Electron.meta_record_map, keys))

        self._metadata = ElectronMeta(session, record, fields=fields)
        self._assets = {}
        self._electron_id = record.id

        self.node_id = record.transport_graph_node_id

    @property
    def query_keys(self) -> list:
        return self._keys

    @property
    def metadata(self) -> ElectronMeta:
        return self._metadata

    @property
    def computed_fields(self) -> Dict:
        return {"sub_dispatch_id": resolve_sub_dispatch_id}

    @property
    def assets(self):
        return self._assets

    @classmethod
    def meta_record_map(cls: DispatchedObject, key: str) -> str:
        return _meta_record_map[key]

    def get_value(self, key: str, session: Session = None, refresh: bool = True):
        return get_filters[key](super().get_value(key, session, refresh))

    def set_value(self, key: str, val: Any, session: Session = None) -> None:
        super().set_value(key, set_filters[key](val), session)


def resolve_sub_dispatch_id(obj: Electron, session: Session) -> str:
    stmt = select(models.Lattice.dispatch_id).where(models.Lattice.electron_id == obj._electron_id)
    record = session.scalars(stmt).first()
    return record



#!/usr/bin/env python

# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Covalent CLI Tool."""

from importlib import metadata
from platform import machine, python_version, system

import click
from rich.console import Console

from .groups import db, deploy
from .service import (
    cluster,
    config,
    logs,
    migrate_legacy_result_object,
    print_header,
    purge,
    restart,
    start,
    status,
    stop,
)


# Main entrypoint
@click.group(invoke_without_command=True)
@click.option("-v", "--version", is_flag=True, help="Display version information.")
@click.pass_context
def cli(ctx: click.Context, version: bool) -> None:
    """
    The Covalent CLI is used to manage and configure Covalent servers.
    """

    console = Console()

    if version:
        print_header(console)
        console.print("Copyright (C) 2021 Agnostiq Inc.", highlight=False)
        console.print(
            f"Using Python {python_version()} on {system()}-{machine()}", highlight=False
        )
        console.print(f"Release version {metadata.version('covalent')}", highlight=False)
    elif ctx.invoked_subcommand is None:
        # Display the help menu if no command was provided
        ctx = click.get_current_context()
        console.print(ctx.get_help())


# Server management
cli.add_command(start)
cli.add_command(stop)
cli.add_command(restart)
cli.add_command(status)
cli.add_command(purge)
cli.add_command(logs)
cli.add_command(cluster)
cli.add_command(db)
cli.add_command(config)
cli.add_command(migrate_legacy_result_object)
cli.add_command(deploy)

if __name__ == "__main__":
    cli()



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Callable, Tuple

from furl import furl

from ..._shared_files import logger
from .. import File
from .transfer_strategy_base import FileTransferStrategy

app_log = logger.app_log


class Blob(FileTransferStrategy):
    """Implements FileTransferStrategy class to transfer files to/from Azure Blob Storage.

    Args:
        client_id: ID of a service principal authorized to perform the transfer
        client_secret: Corresponding secret key for the service principal credentials
        tenant_id: The Azure Active Directory tenant ID which owns the cloud resources.

    Attributes:
        credentials: A tuple containing (client_id, client_secret, tenant_id)
    """

    def __init__(
        self,
        client_id: str = None,
        client_secret: str = None,
        tenant_id: str = None,
    ):
        self.credentials = (tenant_id, client_id, client_secret)

    def _get_blob_service_client(self, storage_account_url):
        """Returns the service client object for the Blob storage account.

        Args:
            storage_account_url: Fully qualified URL of the storage account, e.g., https://account_name.blob.core.windows.net

        Returns:
            blob_service_client: A client object for the Azure Blob service
        """

        try:
            from azure.identity import ClientSecretCredential
            from azure.storage.blob import BlobServiceClient
        except ImportError:
            print(
                "Missing required dependency. Reinstall package as `pip install covalent[azure]`."
            )
            raise

        return BlobServiceClient(
            account_url=storage_account_url, credential=ClientSecretCredential(*self.credentials)
        )

    def _parse_blob_uri(self, blob_uri: str) -> Tuple[str, str, str]:
        """Parses a blob URI and returns the account name, container name, and blob name.

        Args:
            blob_uri: A URI for an Azure Blob object in the form https://<storage_account_name>.blob.core.windows.net/<container_name>/<blob_name>

        Returns:
            parsed_uri: A tuple containing the storage account name, container name, and blob name
        """

        blob_furl = furl(blob_uri)

        storage_account_url = blob_furl.host
        blob_path = blob_furl.path

        storage_container_name = blob_path.segments[0]
        base_path = "/".join(blob_path.segments[1:])

        return (storage_account_url, storage_container_name, base_path)

    def _download_file(self, container_client, blob_name: str, destination_path: Path) -> None:
        """Downloads a single blob to the local filesystem.

        Args:
            container_client: Azure Blob storage container client object
            blob_name: Name of the blob object within the scope of the container
            destination_path: Path on the local filesystem where the file will be saved
        """

        blob_client = container_client.get_blob_client(blob=blob_name)

        with open(destination_path, "wb") as f:
            stream = blob_client.download_blob()
            f.write(stream.readall())

    def download(self, from_file: File, to_file: File = File()) -> Callable:
        """Download files or the contents of folders from Azure Blob Storage.

        Args:
            from_file: File object referencing an object in Azure Blob storage
            to_file: File object referencing a path in the local filesystem

        Returns:
            callable: Download function that is injected into wrapper_fn
        """

        from_filepath = from_file.filepath
        to_filepath = to_file.filepath

        storage_account_url, storage_container_name, base_path = self._parse_blob_uri(
            from_file.uri
        )

        app_log.debug(
            f"Blob download; storage account: {storage_account_url}, from_filepath: {from_filepath}, to_filepath {to_filepath}."
        )

        def callable():
            """Download file or directory from an Azure Blob storage container."""
            from pathlib import Path

            blob_service_client = self._get_blob_service_client(storage_account_url)
            container_client = blob_service_client.get_container_client(storage_container_name)

            blobs = (
                container_client.list_blobs(name_starts_with=base_path)
                if from_file.is_dir
                else [base_path]
            )
            for blob in blobs:
                blob_name = blob.name if from_file.is_dir else base_path
                dest_obj_path = (
                    Path(to_filepath) / Path(blob.name).relative_to(base_path)
                    if to_file.is_dir
                    else Path(to_filepath)
                )
                dest_obj_path.parent.mkdir(parents=True, exist_ok=True)
                self._download_file(container_client, blob_name, dest_obj_path)

        return callable

    def _upload_file(self, container_client, file_path: str, dest_obj_path: Path) -> None:
        """Uploads a single file to Azure Blob storage.

        Args:
            container_client: Azure Blob storage container client object
            file_path: Path on the local filesystem to a file that is uploaded
            dest_obj_path: Path in a blob storage container where the file will be saved
        """

        with open(file_path, "rb") as f:
            container_client.upload_blob(name=str(dest_obj_path), data=f, overwrite=True)

    def upload(self, from_file: File, to_file: File = File()) -> Callable:
        """Upload files or the contents of folders to Azure Blob Storage.

        Args:
            from_file: File object referencing a path in the local filesystem
            to_file: File object referencing an object in Azure Blob storage

        Returns:
            callable: Upload function that is injected into wrapper_fn
        """

        from_filepath = from_file.filepath
        to_filepath = to_file.filepath

        storage_account_url, storage_container_name, base_path = self._parse_blob_uri(to_file.uri)

        app_log.debug(
            f"Blob upload; storage account: {storage_account_url}, from_filepath: {from_filepath}, to_filepath {to_filepath}."
        )

        def callable():
            """Upload file or directory to an Azure Blob storage container."""
            from pathlib import Path

            blob_service_client = self._get_blob_service_client(storage_account_url)
            container_client = blob_service_client.get_container_client(storage_container_name)

            files = (
                [path for path in Path(from_filepath).rglob("*") if path.is_file()]
                if from_file.is_dir
                else [from_filepath]
            )
            for file in files:
                dest_obj_path = (
                    Path(base_path) / Path(file).relative_to(from_filepath)
                    if to_file.is_dir
                    else Path(base_path)
                )
                self._upload_file(container_client, file, dest_obj_path)

        return callable

    def cp(self, from_file: File, to_file: File = File()) -> File:  # pragma: no cover
        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform

from pydantic import BaseModel


class PlatformMetadata(BaseModel):
    """Information about the platform used to run the benchmarks"""

    arch: str = platform.architecture()
    system: str = platform.system()
    machine: str = platform.machine()
    os: str = platform.node()
    python_version: str = platform.python_version()


class PerformanceMetrics(BaseModel):
    """Performance metrics to be gathered"""

    workflow_runtime: float
    covalent_runtime: float
    covalent_speedup: float
    covalent_overhead: float
    covalent_disk_io_time: float = 0.0
    covalent_network_io_time: float = 0.0
    covalent_fraction_disk_io: float = 0.0
    covalent_fraction_idle: float = 0.0
    covalent_fraction_user_mode: float = 0.0
    covalent_fraction_system_mode: float = 0.0
    covalent_dispatch_latency: float = 0.0
    covalent_dispatch_throughput: float = 0.0
    covalent_total_db_reads: float = 0.0
    covalent_total_db_writes: float = 0.0
    covalent_electron_throughput: float = 0.0
    covalent_electron_latency: float = 0.0


class WorkflowBenchmarkResult(BaseModel):
    """Base object representing the results from a single workflow benchmark run"""

    run_id: int
    workflow_name: str
    metadata: PlatformMetadata = PlatformMetadata()
    metrics: PerformanceMetrics



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for defining a local executor that directly invokes the input python function.

This is a plugin executor module; it is loaded if found and properly structured.
"""

import asyncio
import os
from concurrent.futures import ProcessPoolExecutor
from enum import Enum
from typing import Any, Callable, Dict, List, Optional

import requests
from pydantic import BaseModel

from covalent._shared_files import TaskCancelledError, TaskRuntimeError, logger
from covalent._shared_files.config import get_config
from covalent._shared_files.util_classes import RESULT_STATUS, Status
from covalent._shared_files.utils import format_server_url
from covalent.executor import BaseExecutor

# Relative imports are not allowed in executor plugins
from covalent.executor.schemas import ResourceMap, TaskSpec, TaskUpdate

# Store the wrapper function in an external module to avoid module
# import errors during pickling
from covalent.executor.utils.wrappers import io_wrapper, run_task_group

# The plugin class name must be given by the executor_plugin_name attribute:
EXECUTOR_PLUGIN_NAME = "LocalExecutor"

app_log = logger.app_log
log_stack_info = logger.log_stack_info

_EXECUTOR_PLUGIN_DEFAULTS = {
    "log_stdout": "stdout.log",
    "log_stderr": "stderr.log",
    "cache_dir": os.path.join(
        os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"), "covalent"
    ),
    "workdir": os.path.join(
        os.environ.get("XDG_CACHE_HOME") or os.path.join(os.environ["HOME"], ".cache"),
        "covalent",
        "workdir",
    ),
    "create_unique_workdir": False,
}


proc_pool = ProcessPoolExecutor()


# Valid terminal statuses
class StatusEnum(str, Enum):
    CANCELLED = str(RESULT_STATUS.CANCELLED)
    COMPLETED = str(RESULT_STATUS.COMPLETED)
    FAILED = str(RESULT_STATUS.FAILED)


class ReceiveModel(BaseModel):
    status: StatusEnum


MANAGED_EXECUTION = True


class LocalExecutor(BaseExecutor):
    """
    Local executor class that directly invokes the input function.
    """

    SUPPORTS_MANAGED_EXECUTION = MANAGED_EXECUTION

    def __init__(
        self, workdir: str = "", create_unique_workdir: Optional[bool] = None, *args, **kwargs
    ) -> None:
        if not workdir:
            try:
                workdir = get_config("executors.local.workdir")
            except KeyError:
                workdir = _EXECUTOR_PLUGIN_DEFAULTS["workdir"]
                debug_msg = f"Couldn't find `executors.local.workdir` in config, using default value {workdir}."
                app_log.debug(debug_msg)

        if create_unique_workdir is None:
            try:
                create_unique_workdir = get_config("executors.local.create_unique_workdir")
            except KeyError:
                create_unique_workdir = _EXECUTOR_PLUGIN_DEFAULTS["create_unique_workdir"]
                debug_msg = f"Couldn't find `executors.local.create_unique_workdir` in config, using default value {create_unique_workdir}."
                app_log.debug(debug_msg)

        super().__init__(*args, **kwargs)

        self.workdir = workdir
        self.create_unique_workdir = create_unique_workdir

    def run(self, function: Callable, args: List, kwargs: Dict, task_metadata: Dict) -> Any:
        """
        Execute the function locally

        Arg(s)
            function: Function to be executed
            args: Arguments passed to the function
            kwargs: Keyword arguments passed to the function
            task_metadata: Metadata of the task to be executed

        Return(s)
            Task output
        """

        app_log.debug(f"Running function {function} locally")

        self.set_job_handle(42)

        if self.get_cancel_requested():
            app_log.debug("Task has been cancelled don't proceed")
            raise TaskCancelledError

        dispatch_id = task_metadata["dispatch_id"]
        node_id = task_metadata["node_id"]

        if self.create_unique_workdir:
            current_workdir = os.path.join(self.workdir, dispatch_id, f"node_{node_id}")
        else:
            current_workdir = self.workdir

        # Run the target function in a separate process
        fut = proc_pool.submit(io_wrapper, function, args, kwargs, current_workdir)

        output, worker_stdout, worker_stderr, tb = fut.result()

        print(worker_stdout, end="", file=self.task_stdout)
        print(worker_stderr, end="", file=self.task_stderr)

        if tb:
            print(tb, end="", file=self.task_stderr)
            raise TaskRuntimeError(tb)

        return output

    def _send(
        self,
        task_specs: List[TaskSpec],
        resources: ResourceMap,
        task_group_metadata: dict,
    ):
        dispatch_id = task_group_metadata["dispatch_id"]
        task_ids = task_group_metadata["node_ids"]
        gid = task_group_metadata["task_group_id"]
        output_uris = []
        for node_id in task_ids:
            result_uri = os.path.join(self.cache_dir, f"result_{dispatch_id}-{node_id}.pkl")
            stdout_uri = os.path.join(self.cache_dir, f"stdout_{dispatch_id}-{node_id}.txt")
            stderr_uri = os.path.join(self.cache_dir, f"stderr_{dispatch_id}-{node_id}.txt")
            qelectron_db_uri = os.path.join(
                self.cache_dir, f"qelectron_db_{dispatch_id}-{node_id}.mdb"
            )
            output_uris.append((result_uri, stdout_uri, stderr_uri, qelectron_db_uri))

        server_url = format_server_url()

        app_log.debug(f"Running task group {dispatch_id}:{task_ids}")
        future = proc_pool.submit(
            run_task_group,
            list(map(lambda t: t.model_dump(), task_specs)),
            output_uris,
            self.cache_dir,
            task_group_metadata,
            server_url,
        )

        def handle_cancelled(fut):
            app_log.debug(f"In done callback for {dispatch_id}:{gid}, future {fut}")
            if fut.cancelled():
                for task_id in task_ids:
                    url = f"{server_url}/api/v2/dispatches/{dispatch_id}/electrons/{task_id}/job"
                    requests.put(url, json={"status": "CANCELLED"})

        future.add_done_callback(handle_cancelled)

    def _receive(self, task_group_metadata: Dict, data: Any) -> List[TaskUpdate]:
        # Returns (output_uri, stdout_uri, stderr_uri,
        # exception_raised)

        # Job should have reached a terminal state by the time this is invoked.
        dispatch_id = task_group_metadata["dispatch_id"]
        task_ids = task_group_metadata["node_ids"]

        task_results = []

        for task_id in task_ids:
            # Handle the case where the job was cancelled before the task started running
            app_log.debug(f"Receive called for task {dispatch_id}:{task_id} with data {data}")

            if not data:
                terminal_status = RESULT_STATUS.CANCELLED
            else:
                received = ReceiveModel.model_validate(data)
                terminal_status = Status(received.status.value)

            # Don't update any asset metadata since all assets will be
            # pushed from the executor
            task_result = {
                "dispatch_id": dispatch_id,
                "node_id": task_id,
                "status": terminal_status,
                "assets": {},
            }

            task_results.append(TaskUpdate(**task_result))

        app_log.debug(f"Returning results for tasks {dispatch_id}:{task_ids}")
        return task_results

    async def send(
        self,
        task_specs: List[TaskSpec],
        resources: ResourceMap,
        task_group_metadata: dict,
    ):
        # Assets are assumed to be accessible by the compute backend
        # at the provided URIs

        # The Asset Manager is responsible for uploading all assets
        # Returns a job handle (should be JSONable)

        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            None,
            self._send,
            task_specs,
            resources,
            task_group_metadata,
        )

    async def receive(self, task_group_metadata: Dict, data: Any) -> List[TaskUpdate]:
        # Returns (output_uri, stdout_uri, stderr_uri,
        # exception_raised)

        # Job should have reached a terminal state by the time this is invoked.
        loop = asyncio.get_running_loop()

        return await loop.run_in_executor(
            None,
            self._receive,
            task_group_metadata,
            data,
        )
